'''Categorization functions'''


from .globalVariables import *

from . import (utilityFunctions,
              visualizationFunctions,
              extendedDataFrame,
              clusteringFunctions,
              coreFunctions,
              dataStorage)

from .extendedDataFrame import DataFrame


def calculateTimeSeriesCategorization(df_data, dataName, saveDir, hdf5fileName=None, p_cutoff=0.05, fraction=0.75, constantSignalsCutoff=0., lowValuesToTag=1., lowValuesToTagWith=1., NumberOfRandomSamples=10**5, NumberOfCPUs=4, referencePoint=0, autocorrelationBased=True, calculateAutocorrelations=False, calculatePeriodograms=False, preProcessData=True):
        
    """Time series classification.
    
    Parameters:
        df_data: pandas.DataFrame
            Data to process

        dataName: str
            Data name, e.g. "myData_1"

        saveDir: str
            Path of directories poining to data storage

        hdf5fileName: str, Default None
            Preferred hdf5 file name and location

        p_cutoff: float, Default 0.05
            Significance cutoff signals selection

        fraction: float, Default 0.75
            Fraction of non-zero point in a signal

        constantSignalsCutoff: float, Default 0.
            Parameter to consider a signal constant

        lowValuesToTag: float, Default 1.
            Values below this are considered low

        lowValuesToTagWith: float, Default 1.
            Low values to tag with

        NumberOfRandomSamples: int, Default 10**5
            Size of the bootstrap distribution to generate

        NumberOfCPUs: int, Default 4
            Number of processes allowed to use in calculations
            
        referencePoint: int, Default 0
            Reference point

        autocorrelationBased: boolean, Default True
            Whether Autocorrelation of Frequency based

        calculateAutocorrelations: boolean, Default False
            Whether to recalculate Autocorrelations

        calculatePeriodograms: boolean, Default False
            Whether to recalculate Periodograms

        preProcessData: boolean, Default True
            Whether to preprocess data, i.e. filter, normalize etc.

    Returns:
        None

    Usage:
        calculateTimeSeriesCategorization(df_data, dataName, saveDir)
    """

    print('\n', '-'*70, '\n\tProcessing %s (%s)'%(dataName, 'Periodograms' if not autocorrelationBased else 'Autocorrelations'), '\n', '-'*70)

    if not os.path.exists(saveDir):
        os.makedirs(saveDir)

    if hdf5fileName is None:
        hdf5fileName = saveDir + dataName + '.h5'

    df_data = extendedDataFrame.DataFrame(df_data)

    df_data.columns = df_data.columns.astype(float)

    if preProcessData:
        df_data.filterOutAllZeroSignals(inplace=True)
        df_data.filterOutReferencePointZeroSignals(referencePoint=referencePoint, inplace=True)
        df_data.filterOutFractionZeroSignals(fraction, inplace=True)
        df_data.tagValueAsMissing(inplace=True)
        df_data.tagLowValues(lowValuesToTag, lowValuesToTagWith, inplace=True)
        df_data.removeConstantSignals(constantSignalsCutoff, inplace=True)

    dataStorage.write(df_data, saveDir + dataName + '_df_data_transformed', hdf5fileName=hdf5fileName)

    if not autocorrelationBased:
        calculateAutocorrelations = False
        if not calculatePeriodograms:
            df_dataPeriodograms = dataStorage.read(saveDir + dataName + '_dataPeriodograms', hdf5fileName=hdf5fileName)
            df_randomPeriodograms = dataStorage.read(saveDir + dataName + '_randomPeriodograms', hdf5fileName=hdf5fileName)
        
            if (df_dataPeriodograms is None) or (df_randomPeriodograms is None):
                print('Periodograms of data and the corresponding null distribution not found. Calculating...')
                calculatePeriodograms = True
    else:
        calculatePeriodograms = False
        if not calculateAutocorrelations:
            df_dataAutocorrelations = dataStorage.read(saveDir + dataName + '_dataAutocorrelations', hdf5fileName=hdf5fileName)
            df_randomAutocorrelations = dataStorage.read(saveDir + dataName + '_randomAutocorrelations', hdf5fileName=hdf5fileName)
        
            if (df_dataAutocorrelations is None) or (df_randomAutocorrelations is None):
                print('Autocorrelation of data and the corresponding null distribution not found. Calculating...')
                calculateAutocorrelations = True

    if calculatePeriodograms:
        df_data = dataStorage.read(saveDir + dataName + '_df_data_transformed', hdf5fileName=hdf5fileName)

        print('Calculating null distribution (periodogram) of %s samples...' %(NumberOfRandomSamples))
        df_randomPeriodograms = extendedDataFrame.getRandomPeriodograms(df_data, NumberOfRandomSamples=NumberOfRandomSamples, NumberOfCPUs=NumberOfCPUs, fraction=fraction, referencePoint=referencePoint)

        dataStorage.write(df_randomPeriodograms, saveDir + dataName + '_randomPeriodograms', hdf5fileName=hdf5fileName)

        df_data = dataStorage.read(saveDir + dataName + '_df_data_transformed', hdf5fileName=hdf5fileName)
        df_data = df_data.normalizeSignalsToUnity(referencePoint=referencePoint)

        print('Calculating each Time Series Periodogram...')
        df_dataPeriodograms = extendedDataFrame.getLombScarglePeriodogramOfDataframe(df_data)

        dataStorage.write(df_dataPeriodograms, saveDir + dataName + '_dataPeriodograms', hdf5fileName=hdf5fileName)

    if calculateAutocorrelations:
        df_data = dataStorage.read(saveDir + dataName + '_df_data_transformed', hdf5fileName=hdf5fileName)

        print('Calculating null distribution (autocorrelation) of %s samples...' %(NumberOfRandomSamples))
        df_randomAutocorrelations = extendedDataFrame.getRandomAutocorrelations(df_data, NumberOfRandomSamples=NumberOfRandomSamples, NumberOfCPUs=NumberOfCPUs, fraction=fraction, referencePoint=referencePoint)

        dataStorage.write(df_randomAutocorrelations, saveDir + dataName + '_randomAutocorrelations', hdf5fileName=hdf5fileName)

        df_data = dataStorage.read(saveDir + dataName + '_df_data_transformed', hdf5fileName=hdf5fileName)

        df_data = df_data.normalizeSignalsToUnity(referencePoint=referencePoint)

        print('Calculating each Time Series Autocorrelations...')
        df_dataAutocorrelations = utilityFunctions.runCPUs(NumberOfCPUs, coreFunctions.getAutocorrelationsOfData, [(df_data.iloc[i], df_data.columns.values) for i in range(len(df_data.index))])

        df_dataAutocorrelations = pd.DataFrame(data=df_dataAutocorrelations[1::2], index=df_data.index, columns=df_dataAutocorrelations[0])
        df_dataAutocorrelations.columns = ['Lag ' + str(columnID) for columnID in range(len(df_dataAutocorrelations.columns))]
        dataStorage.write(df_dataAutocorrelations, saveDir + dataName + '_dataAutocorrelations', hdf5fileName=hdf5fileName)

    df_data = dataStorage.read(saveDir + dataName + '_df_data_transformed', hdf5fileName=hdf5fileName)

    if not autocorrelationBased:
        df_classifier = df_dataPeriodograms
        df_randomClassifier = df_randomPeriodograms
        info = 'Periodograms'
    else:
        df_classifier = df_dataAutocorrelations
        df_randomClassifier = df_randomAutocorrelations
        info = 'Autocorrelations'

    df_classifier.sort_index(inplace=True)
    df_data.sort_index(inplace=True)

    if not (df_data.index.values == df_classifier.index.values).all():
        raise ValueError('Index mismatch')
            
    QP = [1.0]
    QP.extend([np.quantile(df_randomClassifier.values.T[i], 1. - p_cutoff,interpolation='lower') for i in range(1,df_classifier.shape[1])])
    print('Quantiles:', list(np.round(QP, 16)), '\n')

    significant_index = np.vstack([df_classifier.values.T[lag] > QP[lag] for lag in range(df_classifier.shape[1])]).T

    print('Calculating spikes cutoffs...')
    spike_cutoffs = extendedDataFrame.getRandomSpikesCutoffs(df_data, p_cutoff, NumberOfRandomSamples=NumberOfRandomSamples)
    print(spike_cutoffs)

    df_data = df_data.normalizeSignalsToUnity(referencePoint=referencePoint)

    if not (df_data.index.values == df_classifier.index.values).all():
        raise ValueError('Index mismatch')

    print('Recording SpikeMax data...')
    max_spikes = df_data.index.values[coreFunctions.getSpikes(df_data.values, np.max, spike_cutoffs)]
    print(len(max_spikes))
    significant_index_spike_max = [(gene in list(max_spikes)) for gene in df_data.index.values]
    lagSignigicantIndexSpikeMax = (np.sum(significant_index.T[1:],axis=0) == 0) * significant_index_spike_max
    dataStorage.write(df_classifier[lagSignigicantIndexSpikeMax], saveDir + dataName +'_selected%s_SpikeMax'%(info), hdf5fileName=hdf5fileName)
    dataStorage.write(df_data[lagSignigicantIndexSpikeMax], saveDir + dataName +'_selectedTimeSeries%s_SpikeMax'%(info), hdf5fileName=hdf5fileName)
            
    print('Recording SpikeMin data...')
    min_spikes = df_data.index.values[coreFunctions.getSpikes(df_data.values, np.min, spike_cutoffs)]
    print(len(min_spikes))
    significant_index_spike_min = [(gene in list(min_spikes)) for gene in df_data.index.values]
    lagSignigicantIndexSpikeMin = (np.sum(significant_index.T[1:],axis=0) == 0) * (np.array(significant_index_spike_max) == 0) * significant_index_spike_min
    dataStorage.write(df_classifier[lagSignigicantIndexSpikeMin], saveDir + dataName +'_selected%s_SpikeMin'%(info), hdf5fileName=hdf5fileName)
    dataStorage.write(df_data[lagSignigicantIndexSpikeMin], saveDir + dataName +'_selectedTimeSeries%s_SpikeMin'%(info), hdf5fileName=hdf5fileName)

    print('Recording Lag%s-Lag%s data...'%(1,df_classifier.shape[1]))
    for lag in range(1,df_classifier.shape[1]):
        lagSignigicantIndex = (np.sum(significant_index.T[1:lag],axis=0) == 0) * (significant_index.T[lag])
        dataStorage.write(df_classifier[lagSignigicantIndex], saveDir + dataName +'_selected%s_LAG%s'%(info,lag), hdf5fileName=hdf5fileName)
        dataStorage.write(df_data[lagSignigicantIndex], saveDir + dataName +'_selectedTimeSeries%s_LAG%s'%(info,lag), hdf5fileName=hdf5fileName)
                
    return None

def clusterTimeSeriesCategorization(dataName, saveDir, numberOfLagsToDraw=3, hdf5fileName=None, 
                                    exportClusteringObjects=False, writeClusteringObjectToBinaries=True, autocorrelationBased=True,
                                    method='weighted', metric='correlation', significance='Elbow'):

    """Visualize time series classification.
    
    Parameters:
        dataName: str
            Data name, e.g. "myData_1"

        saveDir: str
            Path of directories pointing to data storage

        numberOfLagsToDraw: int, Default 3
            First top-N lags (or frequencies) to draw

        hdf5fileName: str, Default None
            HDF5 storage path and name

        exportClusteringObjects: boolean, Default False
            Whether to export clustering objects to xlsx files

        writeClusteringObjectToBinaries: boolean, Default True
            Whether to export clustering objects to binary (pickle) files

        autocorrelationBased: boolean, Default True
            Whether to label to print on the plots

        method: str, Default 'weighted'
            Linkage calculation method

        metric: str, Default 'correlation'
            Distance measure

        significance: str, Default 'Elbow'
            Method for determining optimal number of groups and subgroups

    Returns:
        None

    Usage:
        clusterTimeSeriesClassification('myData_1', '/dir1/dir2/')
    """

    info = 'Autocorrelations' if autocorrelationBased else 'Periodograms'

    if hdf5fileName is None:
        hdf5fileName = saveDir + dataName + '.h5'

    def internal(className):
        print('\n\n%s of Time Series:'%(className)) 
        df_data_selected = dataStorage.read(saveDir + dataName + '_selectedTimeSeries%s_%s'%(info, className), hdf5fileName=hdf5fileName)
        df_classifier_selected = dataStorage.read(saveDir + dataName + '_selected%s_%s'%(info, className), hdf5fileName=hdf5fileName)

        if (df_data_selected is None) or (df_classifier_selected is None):

            print('Selected %s time series not found in %s.'%(className, saveDir + dataName + '.h5'))
            print('Do time series classification first.')

            return 

        print('Creating clustering object.')
        clusteringObject = clusteringFunctions.makeClusteringObject(df_data_selected, df_classifier_selected, method=method, metric=metric, significance=significance)

        if clusteringObject is None:
            print('Error creating clustering object')
            return

        print('Exporting clustering object.')
        if writeClusteringObjectToBinaries:
            dataStorage.write(clusteringObject, saveDir + 'consolidatedGroupsSubgroups/' + dataName + '_%s_%s'%(className,info) + '_GroupsSubgroups')
        
        if exportClusteringObjects:
            clusteringFunctions.exportClusteringObject(clusteringObject, saveDir + 'consolidatedGroupsSubgroups/', dataName + '_%s_%s'%(className,info))

        return

    for lag in range(1,numberOfLagsToDraw + 1):
        internal('LAG%s'%(lag))
            
    internal('SpikeMax')
    internal('SpikeMin')

    return None

def visualizeTimeSeriesCategorization(dataName, saveDir, numberOfLagsToDraw=3, autocorrelationBased=True,xLabel='Time', plotLabel='Transformed Expression',horizontal=False, minNumberOfCommunities=2, communitiesMethod='WDPVG', direction='left', weight='distance'):

    """Visualize time series classification.
    
    Parameters:
        dataName: str
            Data name, e.g. "myData_1"

        saveDir: str
            Path of directories pointing to data storage

        numberOfLagsToDraw: boolean, Default 3
            First top-N lags (or frequencies) to draw

        autocorrelationBased: boolean, Default True
            Whether autocorrelation or frequency based

        xLabel: str, Default 'Time'
            X-axis label

        plotLabel: str, Default 'Transformed Expression'
            Label for the heatmap plot

        
        horizontal: boolean, Default False
            Whether to use horizontal or natural visibility graph. 

        minNumberOfCommunities: int, Default 2
            Number of communities to find depends on the number of splits.
            This parameter is ignored in methods that automatically
            estimate optimal number of communities.

        communitiesMethod: str, Default 'WDPVG'
            String defining the method to use for communitiy detection:
                'Girvan_Newman': edge betweenness centrality based approach

                'betweenness_centrality': reflected graph node betweenness centrality based approach

                'WDPVG': weighted dual perspective visibility graph method (note to also set weight variable)

        direction:str, default 'left'
            The direction that nodes aggregate to communities:
                None: no specific direction, e.g. both sides.

                'left': nodes can only aggregate to the left side hubs, e.g. early hubs

                'right': nodes can only aggregate to the right side hubs, e.g. later hubs

        weight: str, Default 'distance'
            Type of weight for communitiesMethod='WDPVG':
                None: no weighted

                'time': weight = abs(times[i] - times[j])

                'tan': weight = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)

                'distance': weight = A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5
    Returns:
        None

    Usage:
        visualizeTimeSeriesClassification('myData_1', '/dir1/dir2/')
    """

    info = 'Autocorrelations' if autocorrelationBased else 'Periodograms'

    def internal(className):
        print('\n\n%s of Time Series:'%(className)) 

        clusteringObject = dataStorage.read(saveDir + 'consolidatedGroupsSubgroups/' + dataName + '_%s_%s'%(className,info) + '_GroupsSubgroups')

        if clusteringObject is None:
            print('Clustering object not found')
            return
        if len(clusteringObject['linkage']) < 2:
            print('Clustering linkage array has only 1 row')
            return 
        
        print('Plotting Dendrogram with Heatmaps.')
        visualizationFunctions.makeDendrogramHeatmapOfClusteringObject(clusteringObject, saveDir, dataName + '_%s_%sBased'%(className,info), AutocorrNotPeriodogr=autocorrelationBased,xLabel=xLabel, plotLabel=plotLabel,horizontal=horizontal, minNumberOfCommunities=minNumberOfCommunities, communitiesMethod=communitiesMethod, direction=direction, weight=weight)

        return

    for lag in range(1,numberOfLagsToDraw + 1):
        internal('LAG%s'%(lag))
            
    internal('SpikeMax')
    internal('SpikeMin')

    return None



'''Annotations and Enumerations'''


import pymysql
import datetime
import urllib.request
import requests

from .globalVariables import *

from . import utilityFunctions
from . import dataStorage


def internalAnalysisFunction(data, multiCorr, MultipleList,  OutputID, InputID, Species, totalMembers,
                            pValueCutoff, ReportFilterFunction, ReportFilter, TestFunction, HypothesisFunction, FilterSignificant,
                            AssignmentForwardDictionary, AssignmentReverseDictionary, prefix, infoDict):

    """Analysis for Multi-Omics or Single-Omics input list
    The function is used internally and not intended to be used directly by user.
    
    Usage:
        Intended for internal use
    """
    
    listData = data[list(data.keys())[0]]

    #If input data was a list of genes, convert it to pairs with label "Generic"
    if not type(listData[0]) is list:
        listData = [[item, 'Unknown'] for item in listData]
    else:
        if len(listData[0])==1:
            listData = [[item[0], 'Unknown'] for item in listData]

    #Get IDs for each gene
    dataForGeneTranslation = [item[0] if type(item) is list else item for item in listData]
    IDs = GeneTranslation(dataForGeneTranslation, OutputID, ConstantGeneDictionary, InputID = InputID,  Species = Species)[OutputID]

    #Remove Missing IDs
    [item.remove('Missing') if 'Missing' in item else None for item in IDs]

    #Might have put in mixed IDs correspocting to different omics types still, give option to user;
    #Extract the translations, match the ones that intersect for different modes (e.g. RNA/Protein label (last element) must be same)
    membersWithAssociations = {}
    for gene, geneIDs in zip(listData, IDs):

        if len(gene)==4:
            geneKey, _, _, geneOmi = gene
        elif len(gene)==3:
            geneKey, geneOmi, _ = gene
        else:
            geneKey, geneOmi = gene

        if MultipleList:
            geneKey += '_' + geneOmi

        if geneKey in membersWithAssociations.keys():
            labels = membersWithAssociations[geneKey][0]
            if not geneOmi in labels:
                labels.append(geneOmi)
        else:
            membersWithAssociations[geneKey] = [[geneOmi],[]]

        for ID in geneIDs:
            if prefix + ID in list(AssignmentForwardDictionary.keys()):
                for AssignmentID in list(AssignmentForwardDictionary[prefix + ID]):
                    if not AssignmentID in membersWithAssociations[geneKey][1]:
                        membersWithAssociations[geneKey][1].append(AssignmentID)
        if len(membersWithAssociations[geneKey][1])==0:
            membersWithAssociations.pop(geneKey)

    allAssignmentIDs = []
    for thisGeneGOlist in [item[1] for item in list(membersWithAssociations.values())]:
        for AssignmentID in thisGeneGOlist:
            if not AssignmentID in allAssignmentIDs:
                allAssignmentIDs.append(AssignmentID)

    testCats = {}
    for AssignmentID in allAssignmentIDs:
        countsInList = len(membersWithAssociations.keys())
        countsInFamily = multiCorr*len(AssignmentReverseDictionary[AssignmentID])
        countsInMembers = np.sum([AssignmentID in item[1] for item in membersWithAssociations.values()])
        whereGeneHits = [AssignmentID in item[1] for item in membersWithAssociations.values()]
        listOfGenesHit = [[item, membersWithAssociations[item][0]] for item in np.array(list(membersWithAssociations.keys()))[whereGeneHits]]
        testValue = TestFunction(countsInList, countsInFamily, multiCorr*totalMembers, countsInMembers)

        testCats[AssignmentID] = [testValue, [countsInList, countsInFamily, multiCorr*totalMembers, countsInMembers], infoDict[AssignmentID], listOfGenesHit]

    correctedpValues = dict(zip(allAssignmentIDs, HypothesisFunction([item[0] for item in list(testCats.values())], pValueCutoff).T))

    for AssignmentID in allAssignmentIDs:
        testCats[AssignmentID][0] = [testCats[AssignmentID][0], correctedpValues[AssignmentID][1], correctedpValues[AssignmentID][2]]

    ResultsHCct = testCats

    #Length filter
    whatIsFilteredLength = ReportFilterFunction(np.array([item[1][3] for item in list(ResultsHCct.values())]), ReportFilter)

    #Significance filter
    whatIsFilteredSignif = np.array([(item[0][2] if FilterSignificant else True) for item in list(ResultsHCct.values())]).astype(bool)

    #Combined filter
    whatIsFiltered = whatIsFilteredLength * whatIsFilteredSignif

    returning = dict(zip(list(np.array(list(ResultsHCct.keys()), dtype=object)[whatIsFiltered]),list(np.array(list(ResultsHCct.values()), dtype=object)[whatIsFiltered])))

    return {list(data.keys())[0]: returning}

def OBOGODictionary(FileURL="http://purl.obolibrary.org/obo/go/go-basic.obo", ImportDirectly=False, PyIOmicaDataDirectory=None, OBOFile="goBasicObo.txt"):

    """Generate Open Biomedical Ontologies (OBO) Gene Ontology (GO) vocabulary dictionary.
    
    Parameters: 
        FileURL: str, Default "http://purl.obolibrary.org/obo/go/go-basic.obo"
            Provides the location of the Open Biomedical Ontologies (OBO) Gene Ontology (GO) 
            file in case this will be downloaded from the web

        ImportDirectly: boolean, Default False
            Import from URL regardles is the file already exists

        PyIOmicaDataDirectory: str, Default None
            Path of directories to data storage

        OBOFile: str, Default "goBasicObo.txt"
            Name of file to store data in (file will be zipped)

    Returns:
        dictionary
            Dictionary of definitions

    Usage:
        OBODict = OBOGODictionary()
    """

    global ConstantPyIOmicaDataDirectory

    PyIOmicaDataDirectory = ConstantPyIOmicaDataDirectory if PyIOmicaDataDirectory==None else PyIOmicaDataDirectory
    fileGOOBO = os.path.join(PyIOmicaDataDirectory, OBOFile)
    fileGOOBOgz = fileGOOBO + '.gz'

    #import the GO OBO file: we check if the OBO file Exist, if not, attempt to download and create it
    if not os.path.isfile(fileGOOBOgz):
        print("Did Not Find Annotation Files, Attempting to Download...")
        ImportDirectly = True

    if os.path.isfile(fileGOOBO):
        os.remove(fileGOOBO)

    if ImportDirectly:
        if os.path.isfile(fileGOOBOgz):
            os.remove(fileGOOBOgz)

        urllib.request.urlretrieve(FileURL.strip('"'), fileGOOBO)

        if os.path.isfile(fileGOOBO):
            print("Created Annotation Files at ", fileGOOBO)
        else:
            print("Did Not Find Annotation Files, Aborting Process")
            return

        with open(fileGOOBO, 'rb') as fileIn:
            with gzip.open(fileGOOBOgz, 'wb') as fileOut:
                shutil.copyfileobj(fileIn, fileOut)

        print("Compressed local file with GZIP.")

        os.remove(fileGOOBO)

    with gzip.open(fileGOOBOgz, 'r') as tempFile:
        inputFile = tempFile.readlines()

    inputFile = [item.decode() for item in inputFile]

    #Find keys "accessions (id):" and "name:" and "namespace" but extract their corresponding values in a list and map them to their corresponding [Term] positions, 
    #Once the "accessions (id):" and its corresponding "name:" in a list, make an association between them,
    #so you can search this association using the key "accessions (id):" to get the value "name:" and "namespace"
    outDictionary = {}

    for position in np.where([item=='[Term]\n'for item in inputFile])[0]:
        
        def getValue(index):

            return inputFile[position + index].strip(['id:', 'name:', 'namespace:'][index - 1]).strip('\n').strip()

        outDictionary[getValue(1)] = [getValue(2), getValue(3)]
    
    return outDictionary

def GetGeneDictionary(geneUCSCTable = None, UCSCSQLString = None, UCSCSQLSelectLabels = None,
                    ImportDirectly = False, Species = "human", KEGGUCSCSplit = [True,"KEGG Gene ID"]):
    
    """Create an ID/accession dictionary from a UCSC search - typically of gene annotations.
    
    Parameters: 
        geneUCSCTable: str, Default None
            Path to a geneUCSCTable file

        UCSCSQLString: str, Default None
            An association to be used to obtain data from the UCSC Browser tables. The key of the association must 
            match the Species option value used (default: human). The value for the species corresponds to the actual MySQL command used

        UCSCSQLSelectLabels: str, Default None
            An association to be used to assign key labels for the data imported from the UCSC Browser tables. 
            The key of the association must match the Species option value used (default: human). The value is a multi component string 
            list corresponding to the matrices in the data file, or the tables used in the MySQL query provided by UCSCSQLString

        ImportDirectly: boolean, Default False
            Import from URL regardles is the file already exists

        Species: str, Default "human"
            Species considered in the calculation, by default corresponding to human

        KEGGUCSCSplit: list, Default [True,"KEGG Gene ID"]
            Two component list, {True/False, label}. If the first component is set to True the initially imported KEGG IDs, 
            identified by the second component label,  are split on + string to fix nomenclature issues, retaining the string following +

    Returns:
        dictionary
            Gene dictionary

    Usage:
        geneDict = GetGeneDictionary()
    """
       
    UCSCSQLSelectLabels = {"human": ["UCSC ID", "UniProt ID", "Gene Symbol", 
        "RefSeq ID", "NCBI Protein Accession", "Ensembl ID", 
        "KEGG Gene ID", "HGU133Plus2 Affymetrix ID"]}

    #Update these to match any change in query
    #Query for UCSC SQL server
    UCSCSQLString = {"human": 
       "SELECT hg19.kgXref.kgID, hg19.kgXref.spID, \
        hg19.kgXref.geneSymbol, hg19.kgXref.refseq, hg19.kgXref.protAcc, \
        hg19.knownToEnsembl.value, hg19.knownToKeggEntrez.keggEntrez, \
        hg19.knownToU133Plus2.value FROM hg19.kgXref LEFT JOIN \
        hg19.knownToEnsembl ON hg19.kgXref.kgID = hg19.knownToEnsembl.name \
        LEFT JOIN hg19.knownToKeggEntrez ON hg19.kgXref.kgID = \
        hg19.knownToKeggEntrez.name LEFT JOIN hg19.knownToU133Plus2 ON \
        hg19.kgXref.kgID = hg19.knownToU133Plus2.name"}

    global ConstantPyIOmicaDataDirectory

    PyIOmicaDataDirectory = ConstantPyIOmicaDataDirectory

    if geneUCSCTable is None:
        geneUCSCTable = os.path.join(PyIOmicaDataDirectory, Species + "GeneUCSCTable" + ".json.gz")
       
    #If the user asked us to import directly, import directly with SQL, otherwise, get it from a directory they specify
    if not os.path.isfile(geneUCSCTable):
        print("Did Not Find Gene Translation Files, Attempting to Download from UCSC...")
        ImportDirectly = True
    else:
        termTable = dataStorage.read(geneUCSCTable, jsonFormat=True)[1]
        termTable = np.array(termTable)

    if ImportDirectly:
        #Connect to the database from UCSC
        ucscDatabase = pymysql.connect("genome-mysql.cse.ucsc.edu","genomep","password")

        if ucscDatabase==None:
            print("Could not establish connection to UCSC. Please try again or add the dictionary manually at ", geneUCSCTable)
            return

        #Prepare a cursor object using cursor() method
        ucscDatabaseCursor = ucscDatabase.cursor()

        try:
            #Execute the SQL command
            ucscDatabaseCursor.execute(UCSCSQLString[Species])

            #Fetch all the rows in a list of lists.
            termTable = ucscDatabaseCursor.fetchall()

        except Exception as exception:
            print(exception)
            print ("Error: unable to fetch data")

        termTable = np.array(termTable).T
        termTable[np.where(termTable=="")] = None

        #Get all the terms we are going to need, import with SQL the combined tables,and export with a time stamp
        dataStorage.write((datetime.datetime.now().isoformat(), termTable.tolist()), geneUCSCTable, jsonFormat=True)

        #Close SQL connection
        ucscDatabase.close()

        if os.path.isfile(geneUCSCTable):
            print("Created Annotation Files at ", geneUCSCTable)
        else:
            print("Did Not Find Annotation Files, Aborting Process")
            return

    returning = {Species : dict(zip(UCSCSQLSelectLabels[Species],termTable))}
    
    if KEGGUCSCSplit[0]:
        returning[Species][KEGGUCSCSplit[1]] = np.array([item.split("+")[1] if item!=None else item for item in returning[Species][KEGGUCSCSplit[1]]])
    
    return returning

def GOAnalysisAssigner(PyIOmicaDataDirectory = None, ImportDirectly = False, BackgroundSet = [], Species = "human",
                        LengthFilter = None, LengthFilterFunction = np.greater_equal, GOFileName = None, GOFileColumns = [2, 5], 
                        GOURL = "http://current.geneontology.org/annotations/"):
    
    """Download and create gene associations and restrict to required background set.

    Parameters: 
        PyIOmicaDataDirectory: str, Default None
            The directory where the default package data is stored

        ImportDirectly: boolean, Default False
            Import from URL regardles is the file already exists

        BackgroundSet: list, Default []
            Background list to create annotation projection to limited background space, involves
            considering pathways/groups/sets and that provides a list of IDs (e.g. gene accessions) that should 
            be considered as the background for the calculation

        Species: str, Default "human"
            Species considered in the calculation, by default corresponding to human

        LengthFilterFunction: function, Default np.greater_equal
            Performs computations of membership in pathways/ontologies/groups/sets, 
            that specifies which function to use to filter the number of members a reported category has 
            compared to the number typically provided by LengthFilter 

        LengthFilter: int, Default None
            Argument for LengthFilterFunction

        GOFileName: str, Default None
            The name for the specific GO file to download from the GOURL if option ImportDirectly is set to True

        GOFileColumns: list, Default [2, 5]
            Columns to use for IDs and GO:accessions respectively from the downloaded GO annotation file, 
            used when ImportDirectly is set to True to obtain a new GO association file

        GOURL: str, Default "http://current.geneontology.org/annotations/"
            The location (base URL) where the GO association annotation files are downloaded from

    Returns:
        dictionary
            Dictionary of IDToGO and GOToID dictionaries

    Usage:
        GOassignment = GOAnalysisAssigner()
    """

    global ConstantPyIOmicaDataDirectory

    PyIOmicaDataDirectory = ConstantPyIOmicaDataDirectory if PyIOmicaDataDirectory==None else PyIOmicaDataDirectory

    #If the user asked us to import PyIOmicaDataDirectoryectly, import PyIOmicaDataDirectoryectly from GO website, otherwise, get it from a PyIOmicaDataDirectoryectory they specify
    file = "goa_" + Species + ".gaf.gz" if GOFileName==None else GOFileName
    localFile =  os.path.join(PyIOmicaDataDirectory, "goa_" + Species + ".gaf")
    localZipFile =  os.path.join(PyIOmicaDataDirectory, "goa_" + Species + ".gaf.gz")
    fileGOAssociations = [os.path.join(PyIOmicaDataDirectory, Species + item + ".json.gz") for item in ["GeneOntAssoc", "IdentifierAssoc"]]

    #We check if the Annotations exist, if not, attempt to download and create them
    if not np.array(list(map(os.path.isfile, fileGOAssociations))).all():
        print("Did Not Find Annotation Files, Attempting to Download...")
        ImportDirectly = True

    if ImportDirectly:
        #Delete existing file
        if os.path.isfile(localFile):
            os.remove(localFile)
        
        urllib.request.urlretrieve(GOURL + file, "\\".join([localZipFile]))
        
        with gzip.open(localZipFile, 'rb') as fileIn:
            with open(localFile, 'wb') as fileOut:
                shutil.copyfileobj(fileIn, fileOut)
        
        #Clean up archive
        os.remove(localZipFile)
        
        with open(localFile, 'r') as tempFile:
            goData = tempFile.readlines()

        #Remove comments by "!" lines
        goData = np.array(goData[np.where(np.array([line[0]!='!' for line in goData]))[0][0]:])
        goData = pd.DataFrame([item.strip('\n').split('\t') for item in goData]).values

        #Remove all entries with missing
        df = pd.DataFrame(goData.T[np.array(GOFileColumns)-1].T)
        df = df[np.count_nonzero(df.values!='', axis=1)==2]

        ##Mark missing values with NaN
        #df.iloc[np.where(df.values=='')] = np.nan

        dfGrouped = df.groupby(df.columns[1])

        keys, values = list(dfGrouped.indices.keys()), list(dfGrouped.indices.values())
        IDs = df.values.T[0]
        geneOntAssoc = dict(zip(keys, [np.unique(IDs[value]).tolist() for value in values]))

        identifierAssoc = utilityFunctions.createReverseDictionary(geneOntAssoc)

        #Save created annotations geneOntAssoc, identifierAssoc
        dataStorage.write((datetime.datetime.now().isoformat(), geneOntAssoc), fileGOAssociations[0], jsonFormat=True)
        dataStorage.write((datetime.datetime.now().isoformat(), identifierAssoc), fileGOAssociations[1], jsonFormat=True)
        
        os.remove(localFile)

        if np.array(list(map(os.path.isfile, fileGOAssociations))).all():
            print("Created Annotation Files at ", fileGOAssociations)
        else:
            print("Did Not Find Annotation Files, Aborting Process")
            return
    else:
        #Otherwise we get from the user specified PyIOmicaDataDirectoryectory
        geneOntAssoc = dataStorage.read(fileGOAssociations[0], jsonFormat=True)[-1]
        identifierAssoc = dataStorage.read(fileGOAssociations[1], jsonFormat=True)[-1]

    if BackgroundSet!=[]:
        #Using provided background list to create annotation projection to limited background space, also remove entries with only one and missing value
        keys, values = np.array(list(identifierAssoc.keys())), np.array(list(identifierAssoc.values()))
        index = np.where([(((len(values[i])==True)*(values[i][0]!=values[i][0]))==False)*(keys[i] in BackgroundSet) for i in range(len(keys))])[0]
        identifierAssoc = dict(zip(keys[index],values[index]))

        #Create corresponding geneOntAssoc
        geneOntAssoc = utilityFunctions.createReverseDictionary(identifierAssoc)

    if LengthFilter!=None:
        keys, values = np.array(list(geneOntAssoc.keys()), dtype=object), np.array(list(geneOntAssoc.values()), dtype=object)
        index = np.where(LengthFilterFunction(np.array([len(value) for value in values]), LengthFilter))[0]
        geneOntAssoc = dict(zip(keys[index],values[index]))

        #Create corresponding identifierAssoc
        identifierAssoc = utilityFunctions.createReverseDictionary(geneOntAssoc)

    return {Species : {"IDToGO": identifierAssoc, "GOToID": geneOntAssoc}}

def obtainConstantGeneDictionary(GeneDictionary, GetGeneDictionaryOptions, AugmentDictionary):
    
    """Obtain gene dictionary - if it exists can either augment with new information or Species or create new, 
    if not exist then create variable.

    Parameters:
        GeneDictionary: dictionary or None
            An existing variable to use as a gene dictionary in annotations. 
            If set to None the default ConstantGeneDictionary will be used

        GetGeneDictionaryOptions: dictionary
            A list of options that will be passed to this internal GetGeneDictionary function

        AugmentDictionary: boolean
            A choice whether or not to augment the current ConstantGeneDictionary global variable or create a new one

    Returns:
        None

    Usage:
        obtainConstantGeneDictionary(None, {}, False)
    """

    global ConstantGeneDictionary
    
    if ConstantGeneDictionary!=None:
        #ConstantGeneDictionary exists
        if AugmentDictionary:
            #Augment ConstantGeneDictionary
            ConstantGeneDictionary = {**ConstantGeneDictionary, **(GetGeneDictionary(**GetGeneDictionaryOptions) if GeneDictionary==None else GeneDictionary)}
        else:
            #Replace ConstantGeneDictionary
            ConstantGeneDictionary = GetGeneDictionary(**GetGeneDictionaryOptions) if GeneDictionary==None else GeneDictionary
    else:
        #Create/load UCSC based translation dictionary - NB global variable or use specified variable
        ConstantGeneDictionary = GetGeneDictionary(**GetGeneDictionaryOptions) if GeneDictionary==None else GeneDictionary

    return None

def GOAnalysis(data, GetGeneDictionaryOptions={}, AugmentDictionary=True, InputID=["UniProt ID","Gene Symbol"], OutputID="UniProt ID",
                 GOAnalysisAssignerOptions={}, BackgroundSet=[], Species="human", OntologyLengthFilter=2, ReportFilter=1, ReportFilterFunction=np.greater_equal,
                 pValueCutoff=0.05, TestFunction=lambda n, N, M, x: 1. - scipy.stats.hypergeom.cdf(x-1, M, n, N), 
                 HypothesisFunction=lambda data, SignificanceLevel: BenjaminiHochbergFDR(data, SignificanceLevel=SignificanceLevel)["Results"], 
                 FilterSignificant=True, OBODictionaryVariable=None,
                 OBOGODictionaryOptions={}, MultipleListCorrection=None, MultipleList=False, GeneDictionary=None):

    """Calculate input data over-representation analysis for Gene Ontology (GO) categories.

    Parameters:
        data: pd.DataFrame or list
            Data to analyze

        GetGeneDictionaryOptions: dictionary, Default {}
            A list of options that will be passed to this internal GetGeneDictionary function

        AugmentDictionary: boolean, Default True
            A choice whether or not to augment the current ConstantGeneDictionary global variable or create a new one

        InputID: list, Default ["UniProt ID","Gene Symbol"]
            Kind of identifiers/accessions used as input

        OutputID: str, Default "UniProt ID"
            Kind of IDs/accessions to convert the input IDs/accession numbers in the function's analysis

        GOAnalysisAssignerOptions: dictionary, Default {}
            A list of options that will be passed to the internal GOAnalysisAssigner function

        BackgroundSet: list, Default []
            Background list to create annotation projection to limited background space, involves
            considering pathways/groups/sets and that provides a list of IDs (e.g. gene accessions) that should be 
            considered as the background for the calculation

        Species: str, Default "human"
            The species considered in the calculation, by default corresponding to human

        OntologyLengthFilter: int, Default 2
            Function that can be used to set the value for which terms to consider in the computation, 
            by excluding GO terms that have fewer items compared to the OntologyLengthFilter value. It is used by the internal
            GOAnalysisAssigner function

        ReportFilter: int, Default 1
            Functions that use pathways/ontologies/groups, and provides a cutoff for membership in ontologies/pathways/groups
            in selecting which terms/categories to return. It is typically used in conjunction with ReportFilterFunction

        ReportFilterFunction: function , Default np.greater_equal
            Specifies what operator form will be used to compare against ReportFilter option value in 
            selecting which terms/categories to return

        pValueCutoff: float, Default 0.05
            Significance cutoff

        TestFunction: function, Default lambda n, N, M, x: 1. - scipy.stats.hypergeom.cdf(x-1, M, n, N)
            Test function

        HypothesisFunction: function, Default lambda data, SignificanceLevel: BenjaminiHochbergFDR(data, SignificanceLevel=SignificanceLevel)["Results"]
            Allows the choice of function for implementing multiple hypothesis testing considerations

        FilterSignificant: boolean, Default True
            Can be set to True to filter data based on whether the analysis result is statistically significant, 
            or if set to False to return all membership computations

        OBODictionaryVariable: str, Default None
            A GO annotation variable. If set to None, OBOGODictionary will be used internally to 
            automatically generate the default GO annotation

        OBOGODictionaryOptions: dictionary, Default {}
            A list of options to be passed to the internal OBOGODictionary function that provides the GO annotations

        MultipleListCorrection: boolean, Default None
            Specifies whether or not to correct for multi-omics analysis. The choices are None, Automatic, 
            or a custom number, e.g protein+RNA

        MultipleList: boolean, Default False
            Specifies whether the input accessions list constituted a multi-omics list input that is annotated so

        GeneDictionary: str, Default None
            Points to an existing variable to use as a gene dictionary in annotations. If set to None 
            the default ConstantGeneDictionary will be used

    Returns:
        dictionary
            Enrichment dictionary

    Usage:
        goExample1 = GOAnalysis(["TAB1", "TNFSF13B", "MALT1", "TIRAP", "CHUK", 
                                "TNFRSF13C", "PARP1", "CSNK2A1", "CSNK2A2", "CSNK2B", "LTBR", 
                                "LYN", "MYD88", "GADD45B", "ATM", "NFKB1", "NFKB2", "NFKBIA", 
                                "IRAK4", "PIAS4", "PLAU"])
    """

    global ConstantGeneDictionary

    #Obtain OBO dictionary with OBOGODictionaryOptions if any. If externally defined use user definition for OBODict Variable
    OBODict = OBOGODictionary(**OBOGODictionaryOptions) if OBODictionaryVariable==None else OBODictionaryVariable

    #Obtain gene dictionary - if it exists can either augment with new information or Species or create new, if not exist then create variable
    obtainConstantGeneDictionary(GeneDictionary, GetGeneDictionaryOptions, AugmentDictionary)
    
    #Get the right GO terms for the BackgroundSet requested and correct Species
    Assignment = GOAnalysisAssigner(BackgroundSet=BackgroundSet, Species=Species , LengthFilter=OntologyLengthFilter) if GOAnalysisAssignerOptions=={} else GOAnalysisAssigner(**GOAnalysisAssignerOptions)
    
    listToggle = False

    #If the input is simply a list
    if type(data) is list:
        data = {'dummy': data}
        listToggle = True

    #The data may be a subgroup from a clustering object, i.e. a pd.DataFrame
    if type(data) is pd.DataFrame:
        id = list(data.index.get_level_values('id'))
        source = list(data.index.get_level_values('source'))
        data = [[id[i], source[i]] for i in range(len(data))]
        data = {'dummy': data}
        listToggle = True

    returning = {}

    #Check if a clustering object
    if "linkage" in data.keys():
        if MultipleListCorrection==None:
            multiCorr = 1
        elif MultipleListCorrection=='Automatic':
            multiCorr = 1
            for keyGroup in sorted([item for item in list(data.keys()) if not item=='linkage']):
                for keySubGroup in sorted([item for item in list(data[keyGroup].keys()) if not item=='linkage']):
                    multiCorr = max(max(np.unique(data[keyGroup][keySubGroup]['data'].index.get_level_values('id'), return_counts=True)[1]), multiCorr)
        else:
            multiCorr = MultipleListCorrection

        #Loop through the clustering object, calculate GO for each SubGroup
        for keyGroup in sorted([item for item in list(data.keys()) if not item=='linkage']):
            returning[keyGroup] = {}
            for keySubGroup in sorted([item for item in list(data[keyGroup].keys()) if not item=='linkage']):
                SubGroupMultiIndex = data[keyGroup][keySubGroup]['data'].index
                SubGroupGenes = list(SubGroupMultiIndex.get_level_values('id'))
                SubGroupMeta = list(SubGroupMultiIndex.get_level_values('source'))
                SubGroupList = [[SubGroupGenes[i], SubGroupMeta[i]] for i in range(len(SubGroupMultiIndex))]

                returning[keyGroup][keySubGroup] = internalAnalysisFunction({keySubGroup:SubGroupList},
                                                                     multiCorr, MultipleList,  OutputID, InputID, Species, len(Assignment[Species]["IDToGO"]),
                                                                     pValueCutoff, ReportFilterFunction, ReportFilter, TestFunction, HypothesisFunction, FilterSignificant,
                                                                     AssignmentForwardDictionary=Assignment[Species]['IDToGO'],
                                                                     AssignmentReverseDictionary=Assignment[Species]['GOToID'],
                                                                     prefix='', infoDict=OBODict)[keySubGroup]

    #The data is a dictionary of type {'Name1': [data1], 'Name2': [data2], ...}
    else:
        for key in list(data.keys()):
            if MultipleListCorrection==None:
                multiCorr = 1
            elif MultipleList and MultipleListCorrection=='Automatic':
                multiCorr = max(np.unique([item[0] for item in data[key]], return_counts=True)[1])
            else:
                multiCorr = MultipleListCorrection

            returning.update(internalAnalysisFunction({key:data[key]}, multiCorr, MultipleList,  OutputID, InputID, Species, len(Assignment[Species]["IDToGO"]),
                                                pValueCutoff, ReportFilterFunction, ReportFilter, TestFunction, HypothesisFunction, FilterSignificant,
                                                AssignmentForwardDictionary=Assignment[Species]['IDToGO'],
                                                AssignmentReverseDictionary=Assignment[Species]['GOToID'],
                                                prefix='', infoDict=OBODict))

        #If a single list was provided, return the association for Gene Ontologies
        returning = returning['dummy'] if listToggle else returning

    return returning

def GeneTranslation(InputList, TargetIDList, GeneDictionary, InputID = None, Species = "human"):

    """Use geneDictionary to convert inputList IDs to different annotations as indicated by targetIDList.
    
    Parameters:
        InputList: list
            List of names

        TargetIDList: list
            Target ID list

        GeneDictionary: dictionary
            An existing variable to use as a gene dictionary in annotations.
            If set to None the default ConstantGeneDictionary will be used

        InputID: str, Default None
            The kind of identifiers/accessions used as input

        Species: str, Default "human"
            The species considered in the calculation, by default corresponding to human
    
    Returns:
        dictionary
            Gene dictionary

    Usage:
        GenDict = GeneTranslation(data, "UniProt ID", ConstantGeneDictionary, InputID = ["UniProt ID","Gene Symbol"],  Species = "human")
    
    """

    if InputID!=None:
        listOfKeysToUse = []
        if type(InputID) is list:
            for key in InputID:
                if key in list(GeneDictionary[Species].keys()):
                    listOfKeysToUse.append(key)
        elif type(InputID) is str:
            listOfKeysToUse.append(InputID)
    else:
        listOfKeysToUse = list(GeneDictionary[Species].keys())
    
    returning = {}

    for TargetID in ([TargetIDList] if type(TargetIDList) is str else TargetIDList):
        returning[TargetID] = {}
        for key in listOfKeysToUse:
            returning[TargetID][key] = []
            for item in InputList:
                allEntries = np.array(GeneDictionary[Species][TargetID])[np.where(np.array(GeneDictionary[Species][key])==item)[0]]
                returning[TargetID][key].append(list(np.unique(allEntries[np.where(allEntries!=None)[0]]) if InputID!=None else allEntries))
                
        #Merge all found lists into one list
        if InputID!=None:
            returningCopy = returning.copy()
            returning[TargetID] = []
            for iitem, item in enumerate(InputList):
                tempList = []
                for key in listOfKeysToUse:
                    tempList.extend(returningCopy[TargetID][key][iitem])
                returning[TargetID].append(tempList)
                    
    return returning

def KEGGAnalysisAssigner(PyIOmicaDataDirectory = None, ImportDirectly = False, BackgroundSet = [], KEGGQuery1 = "pathway", KEGGQuery2 = "hsa",
                        LengthFilter = None, LengthFilterFunction = np.greater_equal, Labels = ["IDToPath", "PathToID"]):

    """Create KEGG: Kyoto Encyclopedia of Genes and Genomes pathway associations, 
    restricted to required background set, downloading the data if necessary.

    Parameters: 
        PyIOmicaDataDirectory: str, Default None
            Directory where the default package data is stored

        ImportDirectly: boolean, Default False
            Import from URL regardles is the file already exists

        BackgroundSet: list, Default []
            A list of IDs (e.g. gene accessions) that should be considered as the background for the calculation

        KEGGQuery1: str, Default "pathway"
            Make KEGG API calls, and sets string query1 in http://rest.kegg.jp/link/<> query1 <> / <> query2. 
            Typically this will be used as the target database to find related entries by using database cross-references

        KEGGQuery2: str, Default "hsa"
            KEGG API calls, and sets string query2 in http://rest.kegg.jp/link/<> query1 <> / <> query2. 
            Typically this will be used as the source database to find related entries by using database cross-references

        LengthFilterFunction: function, Default np.greater_equal
            Option for functions that perform computations of membership in 
            pathways/ontologies/groups/sets, that specifies which function to use to filter the number of members a reported 
            category has compared to the number typically provided by LengthFilter

        LengthFilter: int, Default None
            Allows the selection of how many members each category can have, as typically 
            restricted by the LengthFilterFunction

        Labels: list, Default ["IDToPath", "PathToID"]
            A string list for how keys in a created association will be named

    Returns:
        dictionary
            IDToPath and PathToID dictionary

    Usage:
        KEGGassignment = KEGGAnalysisAssigner()
    """
    
    global ConstantPyIOmicaDataDirectory
    
    PyIOmicaDataDirectory = ConstantPyIOmicaDataDirectory if PyIOmicaDataDirectory==None else PyIOmicaDataDirectory

    ##if the user asked us to import directly, import directly from KEGG website, otherwise, get it from a directory they specify
    fileAssociations = [os.path.join(PyIOmicaDataDirectory, item) for item in [KEGGQuery1 + "_" + KEGGQuery2 + "KEGGMemberToPathAssociation.json.gz", 
                                                                                KEGGQuery1 + "_" + KEGGQuery2 + "KEGGPathToMemberAssociation.json.gz"]]

    if not np.array(list(map(os.path.isfile, fileAssociations))).all():
        print("Did Not Find Annotation Files, Attempting to Download...")
        ImportDirectly = True

    if ImportDirectly:
        localFile = os.path.join(PyIOmicaDataDirectory, KEGGQuery1 + "_" + KEGGQuery2 + ".tsv")
        #Delete existing file
        if os.path.isfile(localFile):
            os.remove(localFile)

        urllib.request.urlretrieve("http://rest.kegg.jp/link/" + KEGGQuery1 + ("" if KEGGQuery2=="" else "/" + KEGGQuery2), localFile)

        with open(localFile, 'r') as tempFile:
            tempLines = tempFile.readlines()
            
        df = pd.DataFrame([line.strip('\n').split('\t') for line in tempLines])

        ##Remove all entries with missing
        #df = pd.DataFrame(goData.T[np.array(GOFileColumns)-1].T)
        #df = df[np.count_nonzero(df.values!='', axis=1)==2]

        dfGrouped = df.groupby(df.columns[1])
        keys, values = list(dfGrouped.indices.keys()), list(dfGrouped.indices.values())
        IDs = df.values.T[0]
        pathToID = dict(zip(keys, [np.unique(IDs[value]).tolist() for value in values]))
        idToPath = utilityFunctions.createReverseDictionary(pathToID)

        dataStorage.write((datetime.datetime.now().isoformat(), idToPath), fileAssociations[0], jsonFormat=True)
        dataStorage.write((datetime.datetime.now().isoformat(), pathToID), fileAssociations[1], jsonFormat=True)

        os.remove(localFile)

        if np.array(list(map(os.path.isfile, fileAssociations))).all():
            print("Created Annotation Files at ", fileAssociations)
        else:
            print("Did Not Find Annotation Files, Aborting Process")
            return
    else:
        #otherwise import the necessary associations from PyIOmicaDataDirectoryectory
        idToPath = dataStorage.read(fileAssociations[0], jsonFormat=True)[1]
        pathToID = dataStorage.read(fileAssociations[1], jsonFormat=True)[1]

    if BackgroundSet!=[]:
        #Using provided background list to create annotation projection to limited background space
        keys, values = np.array(list(idToPath.keys())), np.array(list(idToPath.values()))
        index = np.where([(((len(values[i])==True)*(values[i][0]!=values[i][0]))==False)*(keys[i] in BackgroundSet) for i in range(len(keys))])[0]
        idToPath = dict(zip(keys[index],values[index]))

        #Create corresponding reverse dictionary
        pathToID = utilityFunctions.createReverseDictionary(idToPath)

    if LengthFilter!=None:
        keys, values = np.array(list(pathToID.keys())), np.array(list(pathToID.values()))
        index = np.where(LengthFilterFunction(np.array([len(value) for value in values]), LengthFilter))[0]
        pathToID = dict(zip(keys[index],values[index]))

        #Create corresponding reverse dictionary
        idToPath = utilityFunctions.createReverseDictionary(pathToID)

    return {KEGGQuery2 : {Labels[0]: idToPath, Labels[1]: pathToID}}

def KEGGDictionary(PyIOmicaDataDirectory = None, ImportDirectly = False, KEGGQuery1 = "pathway", KEGGQuery2 = "hsa"):

    """Create a dictionary from KEGG: Kyoto Encyclopedia of Genes and Genomes terms - 
    typically association of pathways and members therein.
    
    Parameters: 
        PyIOmicaDataDirectory: str, Default None
            directory where the default package data is stored

        ImportDirectly: boolean, Default False
            import from URL regardles is the file already exists

        KEGGQuery1: str, Default "pathway"
            make KEGG API calls, and sets string query1 in http://rest.kegg.jp/link/<> query1 <> / <> query2. 
            Typically this will be used as the target database to find related entries by using database cross-references

        KEGGQuery2: str, Default "hsa"
            KEGG API calls, and sets string query2 in http://rest.kegg.jp/link/<> query1 <> / <> query2. 
            Typically this will be used as the source database to find related entries by using database cross-references

    Returns:
        dictionary
            Dictionary of definitions

    Usage:
        KEGGDict = KEGGDictionary()
    """
    
    global ConstantPyIOmicaDataDirectory
    
    PyIOmicaDataDirectory = ConstantPyIOmicaDataDirectory if PyIOmicaDataDirectory==None else PyIOmicaDataDirectory

    #if the user asked us to import directly, import directly from KEGG website, otherwise, get it from a directory they specify
    fileKEGGDict = os.path.join(PyIOmicaDataDirectory, KEGGQuery1 + "_" + KEGGQuery2 + "_KEGGDictionary.json.gz")

    if os.path.isfile(fileKEGGDict):
        associationKEGG = dataStorage.read(fileKEGGDict, jsonFormat=True)[1]
    else:
        print("Did Not Find Annotation Files, Attempting to Download...")
        ImportDirectly = True

    if ImportDirectly:
        queryFile = os.path.join(PyIOmicaDataDirectory, KEGGQuery1 + "_" + KEGGQuery2 + ".tsv")

        if os.path.isfile(queryFile): 
            os.remove(queryFile)

        urllib.request.urlretrieve("http://rest.kegg.jp/list/" + KEGGQuery1 + ("" if KEGGQuery2=="" else "/" + KEGGQuery2), queryFile)

        with open(queryFile, 'r') as tempFile:
            tempLines = tempFile.readlines()
            
        os.remove(queryFile)
        
        associationKEGG = dict([line.strip('\n').split('\t') for line in tempLines])

        dataStorage.write((datetime.datetime.now().isoformat(), associationKEGG), fileKEGGDict, jsonFormat=True)

        if os.path.isfile(fileKEGGDict):
            print("Created Annotation Files at ", fileKEGGDict)
        else:
            print("Did Not Find Annotation Files, Aborting Process")
            return

    return associationKEGG

def KEGGAnalysis(data, AnalysisType = "Genomic", GetGeneDictionaryOptions = {}, AugmentDictionary = True, InputID = ["UniProt ID", "Gene Symbol"],
                OutputID = "KEGG Gene ID", MolecularInputID = ["cpd"], MolecularOutputID = "cpd", KEGGAnalysisAssignerOptions = {}, BackgroundSet = [], 
                KEGGOrganism = "hsa", KEGGMolecular = "cpd", KEGGDatabase = "pathway", PathwayLengthFilter = 2, ReportFilter = 1, 
                ReportFilterFunction = np.greater_equal, pValueCutoff = 0.05, TestFunction = lambda n, N, M, x: 1. - scipy.stats.hypergeom.cdf(x-1, M, n, N), 
                HypothesisFunction = lambda data, SignificanceLevel: BenjaminiHochbergFDR(data, SignificanceLevel=SignificanceLevel)["Results"],
                FilterSignificant = True, KEGGDictionaryVariable = None, KEGGDictionaryOptions = {}, MultipleListCorrection = None, MultipleList = False, 
                GeneDictionary = None, Species = "human", MolecularSpecies = "compound", NonUCSC = False, PyIOmicaDataDirectory = None):

    """Calculate input data over-representation analysis for KEGG: Kyoto Encyclopedia of Genes and Genomes pathways.
    Input can be a list, a dictionary of lists or a clustering object.

    Parameters:
        data: pandas.DetaFrame or list
            Data to analyze

        AnalysisType: str, Default "Genomic"
            Analysis methods that may be used, "Genomic", "Molecular" or "All"

        GetGeneDictionaryOptions: dictionary, Default {}
            A list of options that will be passed to this internal GetGeneDictionary function

        AugmentDictionary: boolean, Default True
            A choice whether or not to augment the current ConstantGeneDictionary global variable or create a new one

        InputID: list, Default ["UniProt ID", "Gene Symbol"]
            The kind of identifiers/accessions used as input

        OutputID: str, Default "KEGG Gene ID"
            A string value that specifies what kind of IDs/accessions to convert the input IDs/accession 
            numbers in the function's analysis

        MolecularInputID: list, Default ["cpd"]
            A string list to indicate the kind of ID to use for the input molecule entries

        MolecularOutputID: str, Default "cpd"
            A string list to indicate the kind of ID to use for the input molecule entries

        KEGGAnalysisAssignerOptions: dictionary, Default {}
            A list of options that will be passed to this internal KEGGAnalysisAssigner function

        BackgroundSet: list, Default []
            A list of IDs (e.g. gene accessions) that should be considered as the background for the calculation

        KEGGOrganism: str, Default "hsa"
            Indicates which organism (org) to use for \"Genomic\" type of analysis (default is human analysis: org=\"hsa\")

        KEGGMolecular: str, Default "cpd"
            Which database to use for molecular analysis (default is the compound database: cpd)

        KEGGDatabase: str, Default "pathway"
            KEGG database to use as the target database

        PathwayLengthFilter: int, Default 2
            Pathways to consider in the computation, by excluding pathways that have fewer items 
            compared to the PathwayLengthFilter value

        ReportFilter: int, Default 1
            Provides a cutoff for membership in ontologies/pathways/groups in selecting which terms/categories 
            to return. It is typically used in conjunction with ReportFilterFunction

        ReportFilterFunction: function, Default np.greater_equal
            Operator form will be used to compare against ReportFilter option value in selecting 
            which terms/categories to return

        pValueCutoff: float, Default 0.05
            A cutoff p-value for (adjusted) p-values to assess statistical significance

        TestFunction: function, Default lambda n, N, M, x: 1. - scipy.stats.hypergeom.cdf(x-1, M, n, N)
            A function used to calculate p-values

        HypothesisFunction: function, Default lambda data, SignificanceLevel: BenjaminiHochbergFDR(data, SignificanceLevel=SignificanceLevel)["Results"]
            Allows the choice of function for implementing multiple hypothesis testing considerations

        FilterSignificant: boolean, Default True
            Can be set to True to filter data based on whether the analysis result is statistically significant, 
            or if set to False to return all membership computations

        KEGGDictionaryVariable: str, Default None
            KEGG dictionary, and provides a KEGG annotation variable. If set to None, KEGGDictionary 
            will be used internally to automatically generate the default KEGG annotation

        KEGGDictionaryOptions: dictionary, Default {}
            A list of options to be passed to the internal KEGGDictionary function that provides the KEGG annotations

        MultipleListCorrection: boolean, Default None
            Specifies whether or not to correct for multi-omics analysis. 
            The choices are None, Automatic, or a custom number

        MultipleList: boolean, Default False 
            Whether the input accessions list constituted a multi-omics list input that is annotated so

        GeneDictionary: str, Default None
            Existing variable to use as a gene dictionary in annotations. If set to None the default ConstantGeneDictionary will be used

        Species: str, Default "human"
            The species considered in the calculation, by default corresponding to human

        MolecularSpecies: str, Default "compound"
            The kind of molecular input

        NonUCSC: , Default 
            If UCSC browser was used in determining an internal GeneDictionary used in ID translations,
            where the KEGG identifiers for genes are number strings (e.g. 4790).The NonUCSC option can be set to True 
            if standard KEGG accessions are used in a user provided GeneDictionary variable, 
            in the form OptionValue[KEGGOrganism] <>:<>numberString, e.g. hsa:4790

        PyIOmicaDataDirectory: str, Default None
            Directory where the default package data is stored

    Returns:
        dictionary
            Enrichment dictionary

    Usage:
        keggExample1 = KEGGAnalysis(["TAB1", "TNFSF13B", "MALT1", "TIRAP", "CHUK", "TNFRSF13C", "PARP1", "CSNK2A1", "CSNK2A2", "CSNK2B", "LTBR", "LYN", "MYD88", 
                                    "GADD45B", "ATM", "NFKB1", "NFKB2", "NFKBIA", "IRAK4", "PIAS4", "PLAU", "POLR3B", "NME1", "CTPS1", "POLR3A"])
    """

    argsLocal = locals().copy()

    global ConstantPyIOmicaDataDirectory

    obtainConstantGeneDictionary(None, {}, True)
    
    PyIOmicaDataDirectory = ConstantPyIOmicaDataDirectory if PyIOmicaDataDirectory==None else PyIOmicaDataDirectory

    #Gene Identifier based analysis
    if AnalysisType=="Genomic":
        #Obtain OBO dictionary. If externally defined use user definition for OBODict Var
        keggDict = KEGGDictionary(**KEGGDictionaryOptions) if KEGGDictionaryVariable==None else KEGGDictionaryVariable

        #Obtain gene dictionary - if it exists can either augment with new information or Species or create new, if not exist then create variable
        obtainConstantGeneDictionary(GeneDictionary, GetGeneDictionaryOptions, AugmentDictionary)

        #get the right KEGG terms for the BackgroundSet requested and correct Species
        Assignment = KEGGAnalysisAssigner(BackgroundSet=BackgroundSet, KEGGQuery1=KEGGDatabase, KEGGQuery2=KEGGOrganism, LengthFilter=PathwayLengthFilter) if KEGGAnalysisAssignerOptions=={} else KEGGAnalysisAssigner(**KEGGAnalysisAssignerOptions)

    #Molecular based analysis
    elif AnalysisType=="Molecular":
        InputID = MolecularInputID
        OutputID = MolecularOutputID
        Species = MolecularSpecies
        NonUCSC = True
        KEGGOrganism = KEGGMolecular
        MultipleListCorrection = None

        keggDict = KEGGDictionary(**({"KEGGQuery1": "pathway", "KEGGQuery2": ""} if KEGGDictionaryOptions=={} else KEGGDictionaryOptions)) if KEGGDictionaryVariable==None else KEGGDictionaryVariable

        #Obtain gene dictionary - if it exists can either augment with new information or Species or create new, if not exist then create variable
        fileMolDict = os.path.join(PyIOmicaDataDirectory, "PyIOmicaMolecularDictionary.json.gz")

        if os.path.isfile(fileMolDict):
            GeneDictionary = dataStorage.read(fileMolDict, jsonFormat=True)[1]
        else:
            fileCSV = os.path.join(PackageDirectory, "data", "MathIOmicaMolecularDictionary.csv")

            print('Attempting to read:', fileCSV)

            if os.path.isfile(fileCSV):
                with open(fileCSV, 'r') as tempFile:
                    tempLines = tempFile.readlines()
            
                tempData = np.array([line.strip('\n').replace('"', '').split(',') for line in tempLines]).T
                tempData = {'compound': {'pumchem': tempData[0].tolist(), 'cpd': tempData[1].tolist()}}
                dataStorage.write((datetime.datetime.now().isoformat(), tempData), fileMolDict, jsonFormat=True)
            else:
                print("Could not find annotation file at " + fileMolDict + " Please either obtain an annotation file from mathiomica.org or provide a GeneDictionary option variable.")
                return

            GeneDictionary = dataStorage.read(fileMolDict, jsonFormat=True)[1]

        obtainConstantGeneDictionary(GeneDictionary, {}, AugmentDictionary)

        #Get the right KEGG terms for the BackgroundSet requested and correct Species
        #If no specific options for function use BackgroundSet, Species request, length request
        Assignment = KEGGAnalysisAssigner(BackgroundSet=BackgroundSet, KEGGQuery1=KEGGDatabase, KEGGQuery2=KEGGOrganism , LengthFilter=PathwayLengthFilter) if KEGGAnalysisAssignerOptions=={} else KEGGAnalysisAssigner(**KEGGAnalysisAssignerOptions)
    
    #Gene Identifier and Molecular based analysis done concurrently
    elif AnalysisType=='All':
        argsMolecular = argsLocal.copy()
        argsMolecular['AnalysisType'] = 'Molecular'
        argsGenomic = argsLocal.copy()
        argsGenomic['AnalysisType'] = 'Genomic'

        return {"Molecular": KEGGAnalysis(**argsMolecular), "Genomic": KEGGAnalysis(**argsGenomic)}
    
    #Abort
    else:
        print("AnalysisType %s is not a valid choice."%AnalysisType)

        return

    listToggle = False

    #If the input is simply a list
    if type(data) is list:
        data = {'dummy': data}
        listToggle = True

    #The data may be a subgroup from a clustering object, i.e. a pd.DataFrame
    if type(data) is pd.DataFrame:
        id = list(data.index.get_level_values('id'))
        source = list(data.index.get_level_values('source'))
        data = [[id[i], source[i]] for i in range(len(data))]
        data = {'dummy': data}
        listToggle = True

    returning = {}

    #Check if a clustering object
    if "linkage" in data.keys():
        if MultipleListCorrection==None:
            multiCorr = 1
        elif MultipleListCorrection=='Automatic':
            multiCorr = 1
            for keyGroup in sorted([item for item in list(data.keys()) if not item=='linkage']):
                for keySubGroup in sorted([item for item in list(data[keyGroup].keys()) if not item=='linkage']):
                    multiCorr = max(max(np.unique(data[keyGroup][keySubGroup]['data'].index.get_level_values('id'), return_counts=True)[1]), multiCorr)
        else:
            multiCorr = MultipleListCorrection

        #Loop through the clustering object, calculate GO for each SubGroup
        for keyGroup in sorted([item for item in list(data.keys()) if not item=='linkage']):
            returning[keyGroup] = {}
            for keySubGroup in sorted([item for item in list(data[keyGroup].keys()) if not item=='linkage']):
                SubGroupMultiIndex = data[keyGroup][keySubGroup]['data'].index
                SubGroupGenes = list(SubGroupMultiIndex.get_level_values('id'))
                SubGroupMeta = list(SubGroupMultiIndex.get_level_values('source'))
                SubGroupList = [[SubGroupGenes[i], SubGroupMeta[i]] for i in range(len(SubGroupMultiIndex))]

                returning[keyGroup][keySubGroup] = internalAnalysisFunction({keySubGroup:SubGroupList},
                                                                             multiCorr, MultipleList,  OutputID, InputID, Species, len(Assignment[KEGGOrganism]["IDToPath"]),
                                                                             pValueCutoff, ReportFilterFunction, ReportFilter, TestFunction, HypothesisFunction, FilterSignificant,
                                                                             AssignmentForwardDictionary=Assignment[KEGGOrganism]['IDToPath'],
                                                                             AssignmentReverseDictionary=Assignment[KEGGOrganism]['PathToID'],
                                                                             prefix='hsa:' if AnalysisType=='Genomic' else '', infoDict=keggDict)[keySubGroup]

    #The data is a dictionary of type {'Name1': [data1], 'Name2': [data2], ...}
    else:
        for key in list(data.keys()):
            if MultipleListCorrection==None:
                multiCorr = 1
            elif MultipleList and MultipleListCorrection=='Automatic':
                multiCorr = max(np.unique([item[0] for item in data[key]], return_counts=True)[1])
            else:
                multiCorr = MultipleListCorrection

            returning.update(internalAnalysisFunction({key:data[key]}, multiCorr, MultipleList,  OutputID, InputID, Species, len(Assignment[KEGGOrganism]["IDToPath"]),
                                                pValueCutoff, ReportFilterFunction, ReportFilter, TestFunction, HypothesisFunction, FilterSignificant,
                                                AssignmentForwardDictionary=Assignment[KEGGOrganism]['IDToPath'],
                                                AssignmentReverseDictionary=Assignment[KEGGOrganism]['PathToID'],
                                                prefix='hsa:' if AnalysisType=='Genomic' else '', infoDict=keggDict))

        #If a single list was provided
        returning = returning['dummy'] if listToggle else returning

    return returning

def MassMatcher(data, accuracy, MassDictionaryVariable = None, MolecularSpecies = "cpd"):

    """Assign putative mass identification to input data based on monoisotopic mass 
    (using PyIOmica's mass dictionary). The accuracy in parts per million. 
    
    Parameters: 
        data: np.array
            Input data

        accuracy: float
            Accuracy

        MassDictionaryVariable: boolean, Default None
            Mass dictionary variable. If set to None, inbuilt 
            mass dictionary (MassDictionary) will be loaded and used

        MolecularSpecies: str, Default "cpd"
            The kind of molecular input

    Returns:
        list
            List of IDs 

    Usage:
       result = MassMatcher(18.010565, 2)
    """
    
    ppm = accuracy*(10**-6)

    MassDict = MassDictionary() if MassDictionaryVariable==None else MassDictionaryVariable
    keys, values = np.array(list(MassDict[MolecularSpecies].keys())), np.array(list(MassDict[MolecularSpecies].values()))

    return keys[np.where((values > data*(1 - ppm)) * (values < data*(1 + ppm)))[0]]

def MassDictionary(PyIOmicaDataDirectory=None):

    """Load PyIOmica's current mass dictionary.
    
    Parameters:
        PyIOmicaDataDirectory: str, Default None
            Directory where the default package data is stored

    Returns:
        dictionary
            Mass dictionary

    Usage:
        MassDict = MassDictionary()
    """

    global ConstantPyIOmicaDataDirectory

    PyIOmicaDataDirectory = ConstantPyIOmicaDataDirectory if PyIOmicaDataDirectory==None else PyIOmicaDataDirectory

    fileMassDict = os.path.join(PyIOmicaDataDirectory, "PyIOmicaMassDictionary.json.gz")

    if os.path.isfile(fileMassDict):
        MassDict = dataStorage.read(fileMassDict, jsonFormat=True)[1]
    else:
        fileCSV = os.path.join(PackageDirectory, "data", "MathIOmicaMassDictionary" +  ".csv")

        if False:
            with open("PyIOmicaMassDictionary", 'r') as tempFile:
                mathDictData = tempFile.readlines()

            mathDict = ''.join([line.strip('\n') for line in mathDictData]).replace('"','').replace(' ','').replace('->',' ').split(',')
            mathDict = np.array([line.split(' ') for line in mathDict])
            np.savetxt(fileCSV, mathDictData, delimiter=',', fmt='%s')

        if os.path.isfile(fileCSV):
            print('Reading:', fileCSV)

            fileMassDictData = np.loadtxt(fileCSV, delimiter=',', dtype=str)
            MassDict = {fileMassDictData[0][0].split(':')[0]: dict(zip(fileMassDictData.T[0],fileMassDictData.T[1].astype(float)))}
            dataStorage.write((datetime.datetime.now().isoformat(), MassDict), fileMassDict, jsonFormat=True)

            print("Created mass dictionary at ", fileMassDict)
        else:
            print("Could not find mass dictionary at ", fileMassDict, 
                    "Please either obtain a mass dictionary file from mathiomica.org or provide a custom file at the above location.")

            return None

    return MassDict

def ExportEnrichmentReport(data, AppendString="", OutputDirectory=None):

    """Export results from enrichment analysis to Excel spreadsheets.
    
    Parameters:
        data: dictionary
            Enrichment results

        AppendString: str, Default ""
            Custom report name, if empty then time stamp will be used

        OutputDirectory: boolean, Default None
            Path of directories where the report will be saved

    Returns:
        None

    Usage:
        ExportEnrichmentReport(goExample1, AppendString='goExample1', OutputDirectory=None)
    """

    def FlattenDataForExport(data):

        returning = {}

        if (type(data) is dict):
            if len(data)==0:
                print('The result is empty.')
                returning['List'] = data
                return 
            idata = data[list(data.keys())[0]]
            if not type(idata) is dict:
                returning['List'] = data
            elif type(idata) is dict:
                idata = idata[list(idata.keys())[0]]
                if not type(idata) is dict:
                    returning = data
                elif type(idata) is dict:
                    idata = idata[list(idata.keys())[0]]
                    if not type(idata) is dict:
                        #Loop through the clustering object
                        for keyClass in list(data.keys()):
                            for keySubClass in list(data[keyClass].keys()):
                                returning[str(keyClass)+' '+str(keySubClass)] = data[keyClass][keySubClass]
                    elif type(idata) is dict:
                        for keyAnalysisType in list(data.keys()):
                            #Loop through the clustering object
                            for keyClass in list(data[keyAnalysisType].keys()):
                                for keySubClass in list(data[keyAnalysisType][keyClass].keys()):
                                    returning[str(keyAnalysisType)+' '+str(keyClass)+' '+str(keySubClass)] = data[keyAnalysisType][keyClass][keySubClass]
        else:
            print('Results type is not supported...')

        return returning

    def ExportToFile(fileName, data):

        writer = pd.ExcelWriter(fileName)

        for key in list(data.keys()):
            keys, values = list(data[key].keys()), list(data[key].values())

            listNum = [[item for sublist in list(value)[:2] for item in sublist] for value in values]
            listNon = [list(value)[2:] for value in values]

            dataDF = [listNum[i] + listNon[i] for i in range(len(keys))]
            columns = ['p-Value', 'BH-corrected p-Value', 'Significant', 'Counts in list', 'Counts in family', 'Total members', 'Counts in members', 'Description', 'List of gene hits']

            df = pd.DataFrame(data=dataDF, index=keys, columns=columns)

            df['Significant'] = df['Significant'].map(bool)

            cleanup = lambda value: value.replace("']], ['", ' | ').replace("[", '').replace("]", '').replace("'", '').replace(", Unknown", '')
            df['List of gene hits'] = df['List of gene hits'].map(str).apply(cleanup)
            df['Description'] = df['Description'].map(str).apply(cleanup)

            df.sort_values(by='BH-corrected p-Value', inplace=True)

            df.to_excel(writer, str(key))

            writer.sheets[str(key)].set_column('A:A', df.index.astype(str).map(len).max()+2)
             
            format = writer.book.add_format({'text_wrap': False,
                                             'valign': 'top'})

            for idx, column in enumerate(df.columns):
                max_len = max((df[column].astype(str).map(len).max(),  # len of largest item
                            len(str(df[column].name)))) + 1            # len of column name/header adding a little extra space

                width = 50 if column=='Description' else min(180, max_len)

                writer.sheets[str(key)].set_column(idx+1, idx+1, width, format)  # set column width

        writer.save()

        print('Saved:', fileName)

        return None
    
    saveDir = os.path.join(os.getcwd(), "Enrichment reports") if OutputDirectory==None else OutputDirectory

    utilityFunctions.createDirectories(saveDir)

    if AppendString=="":
        AppendString=(datetime.datetime.now().isoformat().replace(' ', '_').replace(':', '_').split('.')[0])

    ExportToFile(saveDir + AppendString + '.xlsx', FlattenDataForExport(data))

    return None

def BenjaminiHochbergFDR(pValues, SignificanceLevel=0.05):

    """HypothesisTesting BenjaminiHochbergFDR correction

    Parameters:
        pValues: 1d numpy.array
            Array of p-values

        SignificanceLevel: float, Default 0.05
            Significance level

    Returns:
        dictionary
            Corrected p-Values, p- and q-Value cuttoffs

    Usage:
        result = BenjaminiHochbergFDR(pValues)
    """

    #pValues = np.round(pValues,6)
      
    #count number of hypotheses tested
    nTests = len(pValues)

    #sort the pValues in order
    sortedpVals = np.sort(pValues)

    #generate a sorting ID by ordering
    sortingIDs = np.argsort(np.argsort(pValues))

    #adjust p values to weighted p values
    weightedpVals = sortedpVals * nTests / (1 + np.arange(nTests))

    #flatten the weighted p-values to smooth out any local maxima and get adjusted p-vals
    adjustedpVals = np.array([np.min(weightedpVals[i:]) for i in range(nTests)])

    #finally,generate the qVals by reordering
    qVals = adjustedpVals[sortingIDs]

    ##create an association from which to identify correspondence between
    ##p-values and q-values#Print[{qVals,pValues}];
    pValqValAssociation = dict(zip(qVals, pValues))

    #get the cutoff adjusted q-value
    tempValues = np.flip(adjustedpVals)[np.flip(adjustedpVals) <= SignificanceLevel]
    cutoffqValue = tempValues[0] if len(tempValues) > 0 else np.nan

    #identify corresponding cutoff p-value
    if np.isnan(cutoffqValue):
        cutoffqValue = 0.
        pValCutoff = 0.
    else:
        pValCutoff = pValqValAssociation[cutoffqValue]

    #get q-vals and q-val cutoff, test the qVals for being above or below
    #significance level -- return "true" if enriched and "false" if not
    returning = {"Results": np.vstack((pValues, qVals, qVals <= cutoffqValue)),
                    "p-Value Cutoff": pValCutoff,
                    "q-Value Cutoff": cutoffqValue}

    return returning

def ReactomeAnalysis(data, 
                     uploadURL = 'https://reactome.org/AnalysisService/identifiers/projection',
                     preDownloadURL = 'https://reactome.org/AnalysisService/download/',
                     postDownloadURL = '/pathways/TOTAL/result.csv',
                     headersPOST = {'accept': 'application/json', 'content-type': 'text/plain'},
                     headersGET =  {'accept': 'text/plain'},
                     URLparameters = (('interactors', 'false'), ('pageSize', '20'), ('page', '1'), ('sortBy', 'ENTITIES_PVALUE'), ('order', 'ASC'), ('resource', 'TOTAL'))):

    """Reactome POST-GET-style analysis.
    
    Parameters: 
        data: pd.DataFrame or list
            Data to analyze

        uploadURL: str, Default 'https://reactome.org/AnalysisService/identifiers/projection'
            URL for POST request

        preDownloadURL: str, Default 'https://reactome.org/AnalysisService/download/'
            Part 1 of URL for GET request

        postDownloadURL: str, Default '/pathways/TOTAL/result.csv'
            Part 2 of URL for GET request

        headersPOST: dict, Default {'accept': 'application/json', 'content-type': 'text/plain'}
            URL headers for POST request

        headersGET: dict, Default {'accept': 'text/plain'}
            URL headers for GET request

        URLparameters: tuple, Default (('interactors', 'false'), ('pageSize', '20'), ('page', '1'), ('sortBy', 'ENTITIES_PVALUE'), ('order', 'ASC'), ('resource', 'TOTAL'))
            Parameters for POST request

    Returns:
        returning
            Enrichment object

    Usage:
        goExample1 = ReactomeAnalysis(["TAB1", "TNFSF13B", "MALT1", "TIRAP", "CHUK", 
                                "TNFRSF13C", "PARP1", "CSNK2A1", "CSNK2A2", "CSNK2B", "LTBR", 
                                "LYN", "MYD88", "GADD45B", "ATM", "NFKB1", "NFKB2", "NFKBIA", 
                                "IRAK4", "PIAS4", "PLAU"])
    """
    
    def internalQueryReactome(data):

        '''Function used internally

        Parameters:
            data: str, list or list-like
                Input identifiers in a form of one string, list or a list-like object

        Returns:
            pandas.DataFrame
                Reactome enrichemnt result

        Usage:
            internalQueryReactome('P01023, Q99758, O15439, O43184')
        '''

        # Check or convert data into a string of delimeter-separated values
        if type(data) is str:
            dataString = data.replace("'", "").replace("\"", "")
        else:
            if type(data) is list:
                dataString = str(data).replace("'", "").replace("\"", "").strip(']').strip('[')
            else:
                dataString = str(list(data)).replace("'", "").strip(']').strip('[')

        # POST request to uploadURL with specific 'data'
        response = requests.post(uploadURL, headers=headersPOST, params=URLparameters, data=dataString)

        # Read identifier of the POST request, to use it with GET request
        responseToken = response.json()['summary']['token']

        # GET the csv result from downloadURL using identifier from POST request
        response = requests.get(preDownloadURL + responseToken + postDownloadURL, headers=headersGET)

        # Set up a stream from GET response
        stream = io.StringIO(response.content.decode('utf-8'))
        
        # Read GET response stream into pandas DataFrame
        enrichmentDataFrame = pd.read_csv(stream, index_col=0)

        return enrichmentDataFrame

    listToggle = False

    #If the input is simply a list
    if type(data) is list:
        data = {'dummy': data}
        listToggle = True

    #The data may be a subgroup from a clustering object, i.e. a pd.DataFrame
    if type(data) is pd.DataFrame:
        id = list(data.index.get_level_values('id'))
        data = {'dummy': np.unique(id)}
        listToggle = True
    
    returning = {}

    #Check if a clustering object
    if "linkage" in data.keys():
        #Loop through the clustering object, calculate GO for each SubGroup
        for keyGroup in sorted([item for item in list(data.keys()) if not item=='linkage']):
            returning[keyGroup] = {}
            for keySubGroup in sorted([item for item in list(data[keyGroup].keys()) if not item=='linkage']):
                SubGroupMultiIndex = data[keyGroup][keySubGroup]['data'].index
                SubGroupGenes = list(SubGroupMultiIndex.get_level_values('id'))
                SubGroupList = np.unique(SubGroupGenes)

                returning[keyGroup][keySubGroup] = internalQueryReactome(SubGroupList)
                
    #The data is a dictionary of type {'Name1': [data1], 'Name2': [data2], ...}
    else:
        for key in list(data.keys()):
            returning.update({key: internalQueryReactome(data[key])})

        #If a single list was provided, return the association for Gene Ontologies
        returning = returning['dummy'] if listToggle else returning

    return returning

def ExportReactomeEnrichmentReport(data, AppendString="", OutputDirectory=None):

    """Export results from enrichment analysis to Excel spreadsheets.
    
    Parameters:
        data: dictionary or pandas.DataFrame
            Reactome pathway enrichment results

        AppendString: str, Default ""
            Custom report name, if empty then time stamp will be used

        OutputDirectory: boolean, Default None
            Path of directories where the report will be saved

    Returns:
        None

    Usage:
        ExportReactomeEnrichmentReport(example1, AppendString='example1', OutputDirectory=None)
    """

    def FlattenDataForExport(data):

        returning = {}

        if (type(data) is pd.DataFrame):
            returning['List'] = data
        elif (type(data) is dict):
            idata = data[list(data.keys())[0]]
            if type(idata) is pd.DataFrame:
                returning = data
            elif type(idata) is dict:
                idata = idata[list(idata.keys())[0]]
                if type(idata) is pd.DataFrame:
                    for keyClass in list(data.keys()):
                        for keySubClass in list(data[keyClass].keys()):
                            returning[str(keyClass)+' '+str(keySubClass)] = data[keyClass][keySubClass]
        else:
            print('Results type is not supported...')

        return returning

    def ExportToFile(fileName, data):

        writer = pd.ExcelWriter(fileName)

        for key in list(data.keys()):

            df = data[key]

            df.to_excel(writer, str(key))

            writer.sheets[str(key)].set_column('A:A', df.index.astype(str).map(len).max()+2)
             
            format = writer.book.add_format({'text_wrap': False,
                                             'valign': 'top'})

            for idx, column in enumerate(df.columns):
                max_len = max((df[column].astype(str).map(len).max(),  # len of largest item
                            len(str(df[column].name)))) + 1            # len of column name/header adding a little extra space

                width = 50 if ((column=='Pathway name') or (column=='Found reaction identifiers')) else min(180, max_len)

                writer.sheets[str(key)].set_column(idx+1, idx+1, width, format)  # set column width

        writer.save()

        print('Saved:', fileName)

        return None
    
    saveDir = os.path.join(os.getcwd(), "Enrichment reports") if OutputDirectory==None else OutputDirectory

    utilityFunctions.createDirectories(saveDir)

    if AppendString=="":
        AppendString=(datetime.datetime.now().isoformat().replace(' ', '_').replace(':', '_').split('.')[0])

    ExportToFile(os.path.join(saveDir, AppendString + '.xlsx'), FlattenDataForExport(data))

    return None





'''Data storage functions'''


import h5py
import json
import pickle

from .globalVariables import *

from .extendedDataFrame import DataFrame
from . import utilityFunctions


def read(fileName, withPKLZextension = True, hdf5fileName = None, jsonFormat = False):

    """Read object from a file recorded by function "write". Pandas and Numpy objects are
    read from HDF5 file when provided, otherwise attempt to read from PKLZ file.

    Parameters:
        fileName: str
            Path of directories ending with the file name

        withPKLZextension: boolean, Default True
            Add ".pklz" to a pickle file

        hdf5fileName: str, Default None
            Path of directories ending with the file name. 
            If None then data is read from a pickle file

        jsonFormat: boolean, Default False
            Save data into compressed json file 
    
    Returns:
        data 
            Data object to write into a file

    Usage:
        exampleDataFrame = read('/dir1/exampleDataFrame', hdf5fileName='/dir2/data.h5')
    """

   
    if jsonFormat:
        utilityFunctions.createDirectories("/".join(fileName.split("/")[:-1]))

        with gzip.GzipFile(fileName, 'r') as tempFile:
            data = json.loads(tempFile.read().decode('utf-8'))

        return data

    if hdf5fileName!=None:
        if not os.path.isfile(hdf5fileName):
            print(hdf5fileName, 'not found.')
            return None

        hdf5file = h5py.File(hdf5fileName, 'r')
        
        key = os.path.basename(fileName)
        if key in hdf5file:
            if hdf5file[key].attrs['gtype']=='pd':
                return DataFrame(pd.read_hdf(hdf5fileName, key=key, mode='r'))

        key = 'arrays/' + os.path.basename(fileName)
        if key in hdf5file:
            if hdf5file[key].attrs['gtype']=='np':
                return hdf5file[key].value
        
        searchPickled = print(os.path.basename(fileName), 'not found in', hdf5fileName)

    if hdf5fileName==None or ('searchPickled' in locals()):  
        if not os.path.isfile(fileName + ('.pklz' if withPKLZextension else '')):
            print(fileName + ('.pklz' if withPKLZextension else ''), 'not found.')
            return None

        with gzip.open(fileName + ('.pklz' if withPKLZextension else ''),'rb') as temp_file:
            data = pickle.load(temp_file)

    return data

def write(data, fileName, withPKLZextension = True, hdf5fileName = None, jsonFormat = False):

    """Write object into a file. Pandas and Numpy objects are recorded in HDF5 format
    when 'hdf5fileName' is provided otherwise pickled into a new file.

    Parameters:
        data: any type
            Data object to write into a file

        fileName: str
            Path of directories ending with the file name

        withPKLZextension: boolean, Default True
            Add ".pklz" to a pickle file

        hdf5fileName: str, Default None
            Path of directories ending with the file name. If None then data is pickled

        jsonFormat: boolean, Default False
            Save data into compressed json file 

    Returns:
        None

    Usage:
        write(exampleDataFrame, '/dir1/exampleDataFrame', hdf5fileName='/dir2/data.h5')
    """

    if jsonFormat:
        utilityFunctions.createDirectories("/".join(fileName.split("/")[:-1]))

        with gzip.GzipFile(fileName, 'w') as tempFile:
            tempFile.write(json.dumps(data).encode('utf-8'))

        return None

    if hdf5fileName!=None and type(data) in [pd.DataFrame, DataFrame]:
        utilityFunctions.createDirectories("/".join(hdf5fileName.split("/")[:-1]))
        key=os.path.basename(fileName)

        pd.DataFrame(data).to_hdf(hdf5fileName, key=key, mode='a', complevel=6, complib='zlib')

        hdf5file = h5py.File(hdf5fileName, 'a')
        hdf5file[key].attrs['gtype'] = 'pd'
    elif hdf5fileName!=None and type(data) is np.ndarray:
        utilityFunctions.createDirectories(hdf5fileName)
        hdf5file = h5py.File(hdf5fileName, 'a')
        key = 'arrays/' + fileName.split("/")[-1]
        data = data.astype(float)
        if not key in hdf5file:
            hdf5file.create_dataset(key, data=data, maxshape=tuple([None]*len(data.shape)), dtype=data.dtype,
                                    compression='gzip', compression_opts=6)
        else:
            dataset = hdf5file[key]
            if dataset.shape!=data.shape:
                dataset.resize(data.shape)
            dataset[...] = data
        hdf5file[key].attrs['gtype'] = 'np'
    else:
        if hdf5fileName!=None:
            print('HDF5 format is not supported for data type:', type(data))
            print('Recording data to a pickle file.')

        utilityFunctions.createDirectories("/".join(fileName.split("/")[:-1]))

        with gzip.open(fileName + ('.pklz' if withPKLZextension else ''),'wb') as temp_file:
            pickle.dump(data, temp_file, protocol=4)

    return None




"""
PyIOmica is a general omics data analysis Python package, with a focus on the analysis and categorization of longitudinal datasets.

Usage:
    import pyiomica

Notes:
    For additional information visit: https://github.com/gmiaslab/pyiomica and https://mathiomica.org by G. Mias Lab
"""

print("Loading PyIOmica 1.3.1 (https://github.com/gmiaslab/pyiomica by G. Mias Lab)")


from .globalVariables import *

if printPackageGlobalDefaults:

    variables = locals().copy()

    for variable in variables.items():
        if variable[0][:2] != '__': 
            if not type(variable[1]) is  types.ModuleType:
                if not type(variable[1]) is types.FunctionType:
                    if not type(variable[1]) is types.ClassType:
                        print('\n', variable[0], ':\n', variable[1], '\n')


'''PyIOmica Dataframe extending Pandas DataFrame with new functions'''

import sklearn.preprocessing

from .globalVariables import *

from . import utilityFunctions
from . import coreFunctions


class DataFrame(pd.DataFrame):

    '''Class based on pandas.DataFrame extending capabilities into the doamin of PyIOmica
    
    Initialization parameters are identical to those in pandas.DataFrame
    See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html for detail.    
    '''

    def __init__(self, data=None, index=None, columns=None, dtype=None, copy=False):

        '''Initialization method'''

        super().__init__(data=data, index=index, columns=columns, dtype=dtype, copy=copy)

        return

    def filterOutAllZeroSignals(self, inplace=False):

        """Filter out all-zero signals from a DataFrame.
    
        Parameters:
            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.filterOutAllZeroSignals()

            or

            df_data.filterOutAllZeroSignalse(inplace=True)
       """

        print('Filtering out all-zero signals')

        init = self.shape[0]

        new_data = self.loc[self.index[np.count_nonzero(self, axis=1) > 0]]

        print('Removed ', init - new_data.shape[0], 'signals out of %s.' % init) 
        print('Remaining ', new_data.shape[0], 'signals!')

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

    def filterOutFractionZeroSignals(self, min_fraction_of_non_zeros, inplace=False):
       
        """Filter out fraction-zero signals from a DataFrame.
    
        Parameters:
            min_fraction_of_non_zeros: float
                Maximum fraction of allowed zeros

            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.filterOutFractionZeroSignals(0.75)

            or

            df_data.filterOutFractionZeroSignals(0.75, inplace=True)
       """

        print('Filtering out low-quality signals (with more than %s%% zero points)' %(np.round(100.*(1.-min_fraction_of_non_zeros), 3)))

        min_number_of_non_zero_points = np.int(np.ceil(min_fraction_of_non_zeros * self.shape[1]))
        new_data = self.loc[self.index[np.count_nonzero(self, axis=1) >= min_number_of_non_zero_points]]

        if (self.shape[0] - new_data.shape[0]) > 0:
            print('Removed ', self.shape[0] - new_data.shape[0], 'signals out of %s.'%(self.shape[0])) 
            print('Remaining ', new_data.shape[0], 'signals!')

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return
    
    def filterOutFractionMissingSignals(self, min_fraction_of_non_missing, inplace=False):
       
        """Filter out fraction-zero signals from a DataFrame.
    
        Parameters:
            min_fraction_of_non_missing: float
                Maximum fraction of allowed zeros

            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.filterOutFractionMissingSignals(0.75)

            or

            df_data.filterOutFractionMissingSignals(0.75, inplace=True)
       """

        print('Filtering out low-quality signals (with more than %s%% missing points)' %(np.round(100.*(1.-min_fraction_of_non_missing), 3)))

        min_number_of_non_zero_points = np.int(np.ceil(min_fraction_of_non_missing * self.shape[1]))
        new_data = self.loc[self.index[(~np.isnan(self)).sum(axis=1) >= min_number_of_non_zero_points]]

        if (self.shape[0] - new_data.shape[0]) > 0:
            print('Removed ', self.shape[0] - new_data.shape[0], 'signals out of %s.'%(self.shape[0])) 
            print('Remaining ', new_data.shape[0], 'signals!')

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

    def filterOutReferencePointZeroSignals(self, referencePoint=0, inplace=False):

        """Filter out out first time point zeros signals from a DataFrame.
    
        Parameters:
            referencePoint: int, Default 0
                Index of the reference point
            
            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.filterOutFirstPointZeroSignals()

            or

            df_data.filterOutFirstPointZeroSignals(inplace=True)
       """

        print('Filtering out first time point zeros signals')

        new_data = self.loc[~(self.iloc[:,0] == 0.0)].copy()

        if (self.shape[0] - new_data.shape[0]) > 0:
            print('Removed ', self.shape[0] - new_data.shape[0], 'signals out of %s.'%(self.shape[0])) 
            print('Remaining ', new_data.shape[0], 'signals!')
        
        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return self

    def tagValueAsMissing(self, value=0.0, inplace=False):

        """Tag zero values with NaN.
    
        Parameters:
            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.tagValueAsMissing()

            or

            df_data.tagValueAsMissing(inplace=True)
        """

        print('Tagging %s values with %s'%(value, np.NaN))

        new_data = self.replace(to_replace=value, value=np.NaN, inplace=False)

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

    def tagMissingAsValue(self, value=0.0, inplace=False):

        """Tag NaN with zero.
    
        Parameters:
            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.tagMissingAsValue()

            or

            df_data.tagMissingAsValue(inplace=True)
        """

        print('Tagging %s values with %s'%(np.NaN, value))

        new_data = self.replace(to_replace=np.NaN, value=value, inplace=False)

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

    def tagLowValues(self, cutoff, replacement, inplace=False):

        """Tag low values with replacement value.
    
        Parameters:
            cutoff: float
                Values below the "cutoff" are replaced with "replacement" value

            replacement: float
                Values below the "cutoff" are replaced with "replacement" value
                
            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.tagLowValues(1., 1.)

            or

            df_data.tagLowValues(1., 1., inplace=True)
        """

        print('Tagging low values (<=%s) with %s'%(cutoff, replacement))

        new_data = self.mask(self <= cutoff, other=replacement, inplace=False)

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

    def removeConstantSignals(self, theta_cutoff, inplace=False):

        """Remove constant signals.
      
        Parameters:
            theta_cutoff: float
                Parameter for filtering the signals
                
            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data
            
        Usage:
            df_data = df_data.removeConstantSignals(0.3)

            or

            df_data.removeConstantSignals(0.3, inplace=True)
        """

        print('\nRemoving constant signals. Cutoff value is %s'%(theta_cutoff))

        new_data = self.iloc[np.where(np.std(self,axis=1) / np.mean(np.std(self,axis=1)) > theta_cutoff)[0]]

        print('Removed ', self.shape[0] - new_data.shape[0], 'signals out of %s.' % self.shape[0])
        print('Remaining ', new_data.shape[0], 'signals!')

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

    def boxCoxTransform(self, axis=1, inplace=False):

        """Box-cox transform data.
    
        Parameters:
            axis: int, Default 1
                Direction of processing, columns (1) or rows (0)
            
            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.boxCoxTransformDataframe()

            or

            df_data.boxCoxTransformDataframe(inplace=True)
        """
    
        print('Box-cox transforming raw data')
            
        new_data = self.apply(coreFunctions.boxCoxTransform, axis=axis)

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

    def modifiedZScore(self, axis=0, inplace=False):

        """Z-score (Median-based) transform data.
    
        Parameters:
            axis: int, Default 1
                Direction of processing, rows (1) or columns (0)

            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.modifiedZScoreDataframe()

            or

            df_data.modifiedZScoreDataframe(inplace=True)
        """
            
        print('Z-score (Median-based) transforming box-cox transformed data')

        new_data = self.copy()
        new_data = new_data.apply(coreFunctions.modifiedZScore, axis=axis)

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

    def normalizeSignalsToUnity(self, referencePoint=0, inplace=False):

        """Normalize signals to unity.
    
        Parameters:
            referencePoint: int, Default 0
                Index of the reference point

            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.normalizeSignalsToUnityDataframe()

            or

            df_data.normalizeSignalsToUnityDataframe(inplace=True)
        """

        print('Normalizing signals to unity')

        if not referencePoint is None:
            #Subtract reference time-point value from all time-points
            new_data = self.compareTimeSeriesToPoint(point=referencePoint, inplace=False).copy()
        else:
            new_data = self.copy()

        where_nan = np.isnan(new_data.values.astype(float))
        new_data[where_nan] = 0.0
        new_data = new_data.apply(lambda data: data / np.sqrt(np.dot(data,data)),axis=1)
        new_data[where_nan] = np.nan

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

    def quantileNormalize(self, output_distribution='original', averaging=np.mean, ties=np.mean, inplace=False):

        """Quantile Normalize signals in a DataFrame. 
    
        Note that it is possible there may be equal values within the dataset. In such a scenario, by default, the quantile 
        normalization implementation considered here works by replacing the degenerate values with the mean over all the degenerate ranks.
        Note, that for the default option to work the data should not have any missing values.
        If output_distribution is set to 'uniform' or 'normal' then the scikit-learn's Quantile Transformation is used.

        Parameters:
            output_distribution: str, Default 'original'
                Output distribution. Other options are 'normal' and 'uniform'

            averaging: function, Default np.mean
                With what value to replace the same-rank elements across samples. 
                Default is to take the mean of same-rank elements

            ties: function or str, Default np.mean
                Function or name of the function. How ties should be handled. Default is to replace ties with their mean.
                Other possible options are: 'mean', 'median', 'prod', 'sum', etc.

            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = pd.DataFrame(index=['Gene 1','Gene 2','Gene 3','Gene 4'], columns=['Col 0','Col 1','Col 2'], data=np.array([[5, 4, 3], [2, 1, 4], [3, 4, 6], [4, 2, 8]]))

            df_data = df_data.quantileNormalize()

            or

            df_data.df_data.quantileNormalize(inplace=True)
        """

        print('Quantile normalizing signals...')

        if output_distribution=='original':

            def rankTransform(series, weights):

                se_temp = pd.Series(index=scipy.stats.rankdata(series.values, method='min'), 
                               data=weights[scipy.stats.rankdata(series.values, method='ordinal')-1])

                series[:] = pd.Series(se_temp.index).replace(to_replace=se_temp.groupby(level=0).agg(ties).to_dict()).values

                return series

            weights = averaging(np.sort(self.values, axis=0), axis=1)

            new_data = self.copy()
            new_data = new_data.apply(lambda col: rankTransform(col, weights), axis=0)

        elif output_distribution=='normal' or output_distribution=='uniform': 

            new_data = self.copy()
            new_data.iloc[:] = sklearn.preprocessing.quantile_transform(self.values, output_distribution=output_distribution, n_quantiles=min(self.shape[0],1000), copy=False)

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

    def compareTimeSeriesToPoint(self, point='first', inplace=False):

        """Subtract a particular point of each time series (row) of a Dataframe.
    
        Parameters:     
            point: str, int or float
                Possible options are 'first', 'last', 0, 1, ... , 10, or a value.

            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.compareTimeSeriesToPoint()

            or

            df_data.compareTimeSeriesToPoint(df_data)
        """

        independent = True

        if point == 'first':
            idx = 0
        elif point == 'last':
            idx = len(self.columns) - 1
        elif type(point) is int:
            idx = point
        elif type(point) is float:
            independent = False
        else:
            print("Specify a valid comparison point: 'first', 'last', 0, 1, ..., 10, etc., or a value")
            return

        new_data = self.copy()

        if independent:
            new_data.iloc[:] = (self.values.T - self.values.T[idx]).T
        else:
            new_data.iloc[:] = (self.values.T - point).T

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

    def compareTwoTimeSeries(self, df, function=np.subtract, compareAllLevelsInIndex=True, mergeFunction=np.mean):

        """Create a new Dataframe based on comparison of two existing Dataframes.
    
        Parameters:
            df: pandas.DataFrame
                Data to compare

            function: function, Default np.subtract 
                Other options are np.add, np.divide, or another <ufunc>.

            compareAllLevelsInIndex: boolean, Default True
                Whether to compare all levels in index.
                If False only "source" and "id" will be compared

            mergeFunction: function, Default np.mean
                Input Dataframes are merged with this function, 
                i.e. np.mean (default), np.median, np.max, or another <ufunc>.

        Returns:
            DataFrame or None
                Processed data

        Usage:
            df_data = df_dataH2.compareTwoTimeSeries(df_dataH1, function=np.subtract, compareAllLevelsInIndex=False, mergeFunction=np.median)
        """

        if self.index.names!=df.index.names:
            errMsg = 'Index of Dataframe 1 is not of the same shape as index of Dataframe 2!'
            print(errMsg)
            return errMsg

        if compareAllLevelsInIndex:
            df1_grouped, df2_grouped = self, df
        else:
            def aggregate(df):
                return df.groupby(level=['source', 'id']).agg(mergeFunction)

            df1_grouped, df2_grouped = aggregate(self), aggregate(df)

        index = pd.MultiIndex.from_tuples(list(set(df1_grouped.index.values).intersection(set(df2_grouped.index.values))), 
                                            names=df1_grouped.index.names)

        return function(df1_grouped.loc[index], df2_grouped.loc[index])

    def imputeMissingWithMedian(self, axis=1, inplace=False):

        """Normalize signals to unity.
    
        Parameters:
            axis: int, Default 1
                Axis to apply trasnformation along

            inplace: boolean, Default False
                Whether to modify data in place or return a new one

        Returns:
            Dataframe or None
                Processed data

        Usage:
            df_data = df_data.imputeMissingWithMedian()

            or

            df_data.imputeMissingWithMedian(inplace=True)
        """

        def tempFunction(data):

            data[np.isnan(data)] = np.median(data[np.isnan(data) == False])

            return data

        new_data = self.apply(tempFunction, axis=axis)

        if inplace:
            self._update_inplace(new_data)
        else:
            return self._constructor(new_data).__finalize__(self)

        return

def mergeDataframes(listOfDataframes, axis=0):

    """Merge a list of Dataframes (outer join).
    
    Parameters:
        listOfDataframes: list
            List of pandas.DataFrames

        axis: int, Default 0
            Merge direction. 0 to stack vertically, 1 to stack horizontally

    Returns:
        pandas.Dataframe
            Processed data

    Usage:
        df_data = mergeDataframes([df_data1, df_data2])
    """

    if len(listOfDataframes)==0:
        return None
    elif len(listOfDataframes)==1:
        return listOfDataframes[0]

    df = pd.concat(listOfDataframes, sort=False, axis=axis)

    return DataFrame(df)

def getLombScarglePeriodogramOfDataframe(df_data, NumberOfCPUs=4, parallel=True):

    """Calculate Lomb-Scargle periodogram of DataFrame.
    
    Parameters:
        df: pandas.DataFrame
            Data to process

        parallel: boolean, Default True
            Whether to calculate in parallel mode (>1 process)

        NumberOfCPUs: int, Default 4
            Number of processes to create if parallel is True

    Returns:
        pandas.Dataframe
            Lomb-Scargle periodograms

    Usage:
        df_periodograms = getLombScarglePeriodogramOfDataframe(df_data)
    """

    if parallel:

        results = utilityFunctions.runCPUs(NumberOfCPUs, coreFunctions.pLombScargle, [(series.index[~np.isnan(series)].values, series[~np.isnan(series)].values, df_data.columns.values) for index, series in df_data.iterrows()])

        df_periodograms = pd.DataFrame(data=results[1::2], index=df_data.index, columns=results[0])

    else:
        frequencies = None
        intensities = []

        for index, series in df_data.iterrows():
            values = series[~np.isnan(series)].values
            times = series.index[~np.isnan(series)].values

            tempFrequencies, tempIntensities = coreFunctions.LombScargle(times, values, series.index.values, OversamplingRate=1)

            if frequencies is None:
                frequencies = tempFrequencies

            intensities.append(tempIntensities)

        df_periodograms = pd.DataFrame(data=np.vstack(intensities), index=df_data.index, columns=frequencies)

    return DataFrame(df_periodograms)

def getRandomSpikesCutoffs(df_data, p_cutoff, NumberOfRandomSamples=10**3):

    """Calculate spikes cuttoffs from a bootstrap of provided data,
    gived the significance cutoff p_cutoff.

    Parameters:
        df_data: pandas.DataFrame 
            Data where rows are normalized signals

        p_cutoff: float
            p-Value cutoff, e.g. 0.01

        NumberOfRandomSamples: int, Default 1000
            Size of the bootstrap distribution

    Returns:
        dictionary
            Dictionary of spike cutoffs.

    Usage:
        cutoffs = getSpikesCutoffs(df_data, 0.01)
    """

    data = np.vstack([np.random.choice(df_data.values[:,i], size=NumberOfRandomSamples, replace=True) for i in range(len(df_data.columns.values))]).T

    df_data_random = DataFrame(pd.DataFrame(data=data, index=range(NumberOfRandomSamples), columns=df_data.columns))
    df_data_random.filterOutFractionZeroSignals(0.75, inplace=True)
    df_data_random.normalizeSignalsToUnity(inplace=True)
    df_data_random.removeConstantSignals(0., inplace=True)

    data = df_data_random.values
    counts_non_missing = np.sum(~np.isnan(data), axis=1)
    data[np.isnan(data)] = 0.

    cutoffs = {}

    for i in list(range(data.shape[1]+1)):
        idata = data[counts_non_missing==i]
        if len(idata)>0:
            cutoffs.update({i : (np.quantile(np.max(idata, axis=1), 1.-p_cutoff, interpolation='lower'),
            np.quantile(np.min(idata, axis=1), p_cutoff, interpolation='lower'))} )

    return cutoffs

def getRandomAutocorrelations(df_data, NumberOfRandomSamples=10**5, NumberOfCPUs=4, fraction=0.75, referencePoint=0):

    """Generate autocorrelation null-distribution from permutated data using Lomb-Scargle Autocorrelation.
    NOTE: there should be already no missing or non-numeric points in the input Series or Dataframe

    Parameters:
        df_data: pandas.Series or pandas.Dataframe

        NumberOfRandomSamples: int, Default 10**5
            Size of the distribution to generate

        NumberOfCPUs: int, Default 4
            Number of processes to run simultaneously

    Returns:
        pandas.DataFrame
            Dataframe containing autocorrelations of null-distribution of data.

    Usage:
        result = getRandomAutocorrelations(df_data)
    """

    data = np.vstack([np.random.choice(df_data.values[:,i], size=NumberOfRandomSamples, replace=True) for i in range(len(df_data.columns.values))]).T

    df_data_random = DataFrame(pd.DataFrame(data=data, index=range(NumberOfRandomSamples), columns=df_data.columns))
    df_data_random.filterOutFractionZeroSignals(fraction, inplace=True)
    df_data_random.normalizeSignalsToUnity(inplace=True, referencePoint=referencePoint)
    df_data_random.removeConstantSignals(0., inplace=True)

    print('\nCalculating autocorrelations of %s random samples (sampled with replacement)...'%(df_data_random.shape[0]))

    results = utilityFunctions.runCPUs(NumberOfCPUs, coreFunctions.pAutocorrelation, [(df_data_random.iloc[i].index.values.copy(), 
                                                                                       df_data_random.iloc[i].values.copy(), 
                                                                                       df_data.columns.values.copy()) for i in range(df_data_random.shape[0])])
    
    return pd.DataFrame(data=results[1::2], columns=results[0])

def getRandomPeriodograms(df_data, NumberOfRandomSamples=10**5, NumberOfCPUs=4, fraction=0.75, referencePoint=0):

    """Generate periodograms null-distribution from permutated data using Lomb-Scargle function.

    Parameters:
        df_data: pandas.Series or pandas.Dataframe

        NumberOfRandomSamples: int, Default 10**5
            Size of the distribution to generate

        NumberOfCPUs: int, Default 4
            Number of processes to run simultaneously

    Returns:
        pandas.DataFrame
            Dataframe containing periodograms

    Usage:
        result = getRandomPeriodograms(df_data)
    """

    data = np.vstack([np.random.choice(df_data.values[:,i], size=NumberOfRandomSamples, replace=True) for i in range(len(df_data.columns.values))]).T

    df_data_random = DataFrame(pd.DataFrame(data=data, index=range(NumberOfRandomSamples), columns=df_data.columns))
    df_data_random.filterOutFractionZeroSignals(fraction, inplace=True)
    df_data_random.normalizeSignalsToUnity(inplace=True, referencePoint=referencePoint)
    df_data_random.removeConstantSignals(0., inplace=True)

    print('\nCalculating periodograms of %s random samples (sampled with replacement)...'%(df_data_random.shape[0]))

    return getLombScarglePeriodogramOfDataframe(df_data_random, NumberOfCPUs=NumberOfCPUs)



# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
from scipy.stats.stats import pearsonr
from collections import Counter
from copy import deepcopy

from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score

import matplotlib.pyplot as plt


def bootstrapGeneral(df, N, shuffling = True):
    '''To generate bootstrap samples
    
    Parameters:
        df: pandas dataframe
            the source dataframe using to generate bootstrap samples
        N: integer
            the size of bootstrap samples
        shuffling : boolean
            shuffle the data or not, The default is True.
    
    Returns:
        bootstrapDF: pandas dataframe
            the bootstrap samples
    '''


    bootstrapDF = pd.DataFrame()
    df.reset_index(drop=True, inplace=True)
    if shuffling:
        for column in df:
            bootstrapDF[column] = df[column].iloc[np.random.randint(df.shape[0], size=N)].values
            
    else:
        bootstrapDF = df.iloc[np.random.randint(df.shape[0], size=N)]
    
    return bootstrapDF

def calculateLinksBetweenSubjectsByDistance(df1,df2,cutoff):
    '''To calculate the linked time series/Genes from two dataframes base on the Euclidean distance
    
    Parameters:
        df1: pandas dataframe
            the first time series from df1
        df2: pandas dataframe
            the second time series from df2
        cutoff: float
            if the distance between two time series less than cutoff, the two time series is linked
            time series
    
    Returns:
        numlinkedGenes: integer/float
            number of linked time series
        commonGenes: integer/float
            number of common time series in df1 and df2
        linkedGenes: list of string
            the ids/names of linked time series
    '''
    
    idx = df1.index.intersection(df2.index)
    numlinkedGenes = 0
    linkedGenes = []
    commonGenes = len(idx)
    if commonGenes != 0:
        for r in idx:
            s1 = df1.loc[r]
            s2 = df2.loc[r]
            
            d = np.linalg.norm(s1-s2)
            if d < cutoff:
                numlinkedGenes += 1
                linkedGenes.append(r)
    return numlinkedGenes, commonGenes,linkedGenes

def calculateLinksBetweenSubjectsByCorrelation(df1,df2,cutoff):
    '''To calculate the linked time series/Genes from two dataframes base on the pearson correlation
    
    Parameters:
        df1: pandas dataframe
            the first time series from df1
        df2: pandas dataframe
            the second time series from df2
        cutoff: float
            if the pearson correlation between two time series less than cutoff, the two time series is linked
            time series
    
    Returns:
        numlinkedGenes: integer/float
            number of linked time series
        commonGenes: integer/float
            number of common time series in df1 and df2
        linkedGenes: list of string
            the ids/names of linked time series
    
    '''
    
    idx = df1.index.intersection(df2.index)
    numlinkedGenes = 0
    linkedGenes = []
    commonGenes = len(idx)
    if commonGenes != 0:
        for r in idx:
            s1 = df1.loc[r]
            s2 = df2.loc[r]
            
            d = pearsonr(s1, s2)[0]
            if d > cutoff:
                numlinkedGenes += 1
                linkedGenes.append(r)
    return numlinkedGenes, commonGenes,linkedGenes

def getCommunityStructure(cs):
    '''
    To change community structure from {node1:community1, node2:community2,...} to
    {community1:[node1, node2,...], community2:[node3, node4,...]}

    Parameters:
        cs: dictionary
            the community structure as {node1:community1, node2:community2,...}
        
    Returns:
        community_structure: dictionary
            the community structure as {community1:[node1, node2,...], community2:[node3, node4,...]}
    '''
    commu_list = list(cs.values())
    community_structure = {}
    for commu in commu_list:
        templist = []
        for k,v in cs.items():
            if int(v) == int(commu):
                templist.append(k)
        community_structure[commu] = deepcopy(templist)
    return community_structure

def getCommunityGenesDict(community_structure,genelist,endwithString):
    '''To get gene IDs list of each community within selected individuals' category 

    Parameters:
        community_structure: dictionary
                the community structure as {community1:[node1, node2,...], community2:[node3, node4,...]}
        genelist: dictionary
            the gene list of each individuals, the key is the id of individual 
        endwithString: list of string
           the selected individuals categories, which attached to the end of the individual ids 
    
    Returns:
        community_genes_dict: dictionary
            the genes list of each community
    
    '''

    community_genes_dict = {}
    for key in community_structure.keys():
    #for k in [1]:    
        genes_list = []
        subjectlist = community_structure[key]
        for k in genelist.keys():
            if (k[0].endswith(endwithString) and k[1].endswith(endwithString)
                and k[0] in subjectlist and k[1] in subjectlist):
            #if sub.endswith('Prediabetic'):
                genes_list.append(set(genelist[k]))
            
        #copy_genes_list = deepcopy(genes_list)
        
        if len(genes_list) > 0:
            genes_set = list(set.intersection(*genes_list))
            if len(genes_set):
                community_genes_dict[key] = genes_set
    community_genes_dict = {k: v for k, v in community_genes_dict.items() if len(v)}
    return community_genes_dict

def splitGenes(community_gene_dict):
    '''Split gene ids, to seperate the genes name from attached labels

    Parameters:
        community_gene_dict: dictionary
            the genes ids list of each community
    
    Returns:
        new_dict: dictionary
            the gene names list of each community
    '''
    new_dict = {}
    for key in community_gene_dict:
        fl = []
        l = community_gene_dict[key]
        for name in l:
            fl.append(name.split('+')[0])
        new_dict[key] = fl
        
    return new_dict

def getCommunityTopGenesByNumber(community_structure,genelist,endwithString,numberOfTopGenes=500):
    '''To get the top ranking genes of each community

    Parameters:
        community_structure: dictionary
            the community structure as {community1:[node1, node2,...], community2:[node3, node4,...]}
        genelist: dictionary
            the genes list of each community 
        endwithString: list of string
            the selected individuals categories, which attached to the end of the individual ids 
        numberOfTopGenes: integer, optional
            the number of top ranking genes. The default is 500.
    
    Returns:
        community_genes_dict: dictionary
            the top ranking genes of each community
    '''

    community_genes_dict = {}
    for key in community_structure.keys():
    #for k in [1]:    
        genes_list = []
        subjectlist = community_structure[key]
        for k in genelist.keys():
            if (k[0].endswith(endwithString) and k[1].endswith(endwithString)
                and k[0] in subjectlist and k[1] in subjectlist):
            #if sub.endswith('Prediabetic'):
                genes_list.extend(genelist[k])
            
        counted = Counter(genes_list)
        ordered = [value for value, count in counted.most_common()]
        
        if len(ordered) >= numberOfTopGenes:
            community_genes_dict[key] = ordered[0:numberOfTopGenes]
        else:
            print("number of genes less than choose number of top genes")
            community_genes_dict[key] = ordered
            
    community_genes_dict = {k: v for k, v in community_genes_dict.items() if len(v)}
            
    return community_genes_dict

def getCommunityTopGenesByFrequencyRanking(community_structure,genelist,endwithString,frequencyPercentage=50):
    '''To get the top frequency genes of each community
    
    Parameters:
        community_structure: dictionary
            the community structure as {community1:[node1, node2,...], community2:[node3, node4,...]}
        genelist: dictionary
            the genes list of each community 
        endwithString: list of string
            the selected individuals categories, which attached to the end of the individual ids 
        frequencyPercentage: float, optional
            the top percentage frequency of choosed genes, The default is 50.
    
    Returns:
        community_genes_dict: dictionary
            the top percentage frequency genes of each community
    '''

    community_genes_dict = {}
    for key in community_structure.keys():
    #for k in [1]:    
        genes_list = []
        subjectlist = community_structure[key]
        for k in genelist.keys():
            if (k[0].endswith(endwithString) and k[1].endswith(endwithString)
                and k[0] in subjectlist and k[1] in subjectlist):
            #if sub.endswith('Prediabetic'):
                genes_list.extend(genelist[k])
            
        counted = Counter(genes_list)
        top_ranking_value = round(max(counted.values()) * (1 - frequencyPercentage/100))
        genes = [k for k, v in counted.items() if int(v) >= top_ranking_value]

        community_genes_dict[key] = genes
           
    community_genes_dict = {k: v for k, v in community_genes_dict.items() if len(v)}
            
    return community_genes_dict

def optimizeK(df,rangeK,saveFig=False,**kargs):
    '''To optimize the k value of k-mean cluster

    Parameters:
        df: pandas dataframe
            the data source to do k-mean cluster
        rangeK: python range, e.g. rangeK = range(0,10)
            the K value range
        saveFig: boolean, optional
            save figure or not. The default is False.
        \*\*kargs: figure name
            if saveFig is true, the \*\*kargs is the figure name 
    
    Returns:
        optimizek:integer
            the optimized K value
    
    '''
    
    sil = []
    elbow = []
    K = rangeK
    for k in K:
        km = KMeans(n_clusters=k)
        km = km.fit(df)
        labels = km.labels_
        elbow.append(km.inertia_)
        sil.append(silhouette_score(df, labels, metric = 'euclidean'))
    
    maxindex = sil.index(max(sil))
    optimizek = K[maxindex]
    
    if saveFig:
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 6)) # create figure & 1 axis
        ax1.plot(K, sil, 'bx-')
        ax1.set_xlabel('k')
        ax1.set_ylabel('Silhouette_score')
        ax1.set_title('Silhouette Method For Optimal k')
        
        ax2.plot(K, elbow, 'rx-')
        ax2.set_xlabel('k')
        ax2.set_ylabel('Sum_of_squared_distances')
        ax2.set_title('Elbow Method For Optimal k')
        plt.tight_layout()
        fig.savefig(kargs['figname'])   # save the figure to file
        plt.close(fig) 
    
    return optimizek
    


'''Core function of PyIOmica'''


from .globalVariables import *


def modifiedZScore(subset, printValues=False):

    """Calculate modified z-score based on "Median absolute deviation".

    Parameters:
        subset: pandas.Series
            Data to transform

    Returns:
        pandas.Series
            Transformed subset

    Usage:
        data = modifiedZScore(data)
    """

    def medianAbsoluteDeviation(expr, axis=None):

        """1D, 2D Median absolute deviation of a sequence of numbers or pd.Series.

        Parameters:
            expr: pandas.Series, pandas.DataFrame or numpy.array
                Data for analysis

            axis: int, Default None
                Multidimensional arrays are flattened, 0: use if data in columns, 1: use if data in rows

        Returns:
            float
                Median absolute deviation (M.A.D.)

        Usage:
            MedianAD = medianAbsoluteDeviation(data, axis=None)
        """

        data = None

        if isinstance(expr, np.ndarray):

            data = expr

        elif isinstance(expr, (pd.Series, pd.DataFrame)):

            data = expr.values

        try:

            if len(data) > 1:

                if axis == None or axis == 0:

                    return np.median(np.abs(data - np.median(data,axis)),axis)

                elif axis == 1:

                    if len(data.shape) < 2:

                        print('Warning: axis = %s option is invalid for 1-D array...' % (axis))

                    else:

                        return np.median((np.abs(data.transpose() - np.median(data,axis)).transpose()),axis)

        except Exception as exception:
            print(exception)

            print('Unsupported data type: ', type(expr))
        
        return None

    values = subset[~np.isnan(subset.values)].values

    MedianAD = medianAbsoluteDeviation(values, axis=None)

    if MedianAD == 0.:
        MeanAD = np.sum(np.abs(values - np.mean(values))) / len(values)
        if printValues:
            print('MeanAD:', MeanAD, '\tMedian:', np.median(values))
        coefficient = 0.7978846 / MeanAD
    else:
        if printValues:
            print('MedianAD:', MedianAD, '\tMedian:', np.median(values))
        coefficient = 0.6744897 / MedianAD
        
    subset.iloc[~np.isnan(subset.values)] = coefficient * (values - np.median(values))

    return subset

def boxCoxTransform(subset, lmbda=None, giveLmbda=False, printLambda=True):

    """Power transform from scipy.stats

    Parameters:
        subset: pandas.Series
            Data to transform

        lmbda: float, Default None
            Lambda parameter, if not specified optimal value will be determined

        giveLmbda: boolean, Default False
            Also return Lambda value

        printLmbda: boolean, Default False
            Print Lambda value

    Returns:
        pandas.Series
            Transformed data and Lambda parameter

    Usage:
        myData = boxCoxTransform(myData)
    """

    where_negative = np.where(subset < 0)
    if len(where_negative[0]>0):
        errMsg = 'Warning: negative values are present in the data. Review the sequence of the data processing steps.'
        print(errMsg)

    where_positive = np.where(subset > 0)

    if lmbda == None:
        transformed_data = scipy.stats.boxcox(subset.values[where_positive])
    else:
        transformed_data = (scipy.stats.boxcox(subset.values[where_positive], lmbda=lmbda),lmbda)

    subset.iloc[where_positive] = transformed_data[0]

    lmbda = transformed_data[1]

    if giveLmbda:

        return subset, lmbda

    if printLambda:
        print('Fitted lambda:', lmbda)

    return subset

def ampSquaredNormed(func, freq, times, data):

    """Lomb-Scargle core function. For internal use only.
    Calculate the different frequency components of our spectrum: project the cosine/sine component and normalize it

    Parameters:
        func: function
            One of two functions, i.e. Sin or Cos

        freq: float
            Frequency

        times: 1d numpy.array
            Input times (starting point adjusted w.r.t.dataset times), Zero-padded

        data: 1d numpy.array
            Input Data with the mean subtracted from it, before zero-padding.

    Returns:
        float
            Squared amplitude normalized.

    Usage:
        coef = ampSquaredNormed(np.cos, freguency, inputTimesNormed, inputDataCentered)

        Intended for internal use only.
    """

    omega_freq = 2. * (np.pi) * freq
    theta_freq = 0.5 * np.arctan2(np.sum(np.sin(4. * (np.pi) * freq * times)), np.sum(np.cos(4. * (np.pi) * freq * times) + 10 ** -20))
    
    ampSum = np.sum(data * func(omega_freq * times - theta_freq)) ** 2
    ampNorm = np.sum(func(omega_freq * times - theta_freq) ** 2)

    return chop(ampSum) / ampNorm

def autocorrelation(inputTimes, inputData, inputSetTimes, UpperFrequencyFactor=1):
    
    """Autocorrelation function

    Parameters:
        inputTimes: 1d numpy.array
            Times corresponding to provided data points

        inputData: 1d numpy.array
            Data points

        inputSetTimes: 1d numpy.array
            A complete set of all possible N times during which data could have been collected.

        UpperFrequencyFactor: int, Default 1
            Upper frequency factor.

    Returns:
        2d numpy.array
            Array of time lags with corresponding autocorrelations

    Usage:
        result = autocorrelation(inputTimes, inputData, inputSetTimes)
    """
    # in case the values passed contain nan entries, theses are removed - this could affect window estimation and ensuing functions.
    inputTimes = inputTimes[~np.isnan(inputData)]
    inputData = inputData[~np.isnan(inputData)]
    def InverseAutocovariance(inputTimes, inputData, inputSetTimes, UpperFrequencyFactor=1):

        #adjust inputTimes starting point w.r.t.dataset times, AND ZERO-PAD THEM
        inputTimesNormed = np.concatenate((inputTimes, inputSetTimes + inputSetTimes[-1])) - inputSetTimes[0]

        #calculate the number of timepoints in the overall set-since we cut
        #freqStep to half f0, we should compensate n by multiplication by two
        n = 2 * len(inputSetTimes)

        #calculate the time window of observation
        window = np.max(inputSetTimes) - np.min(inputSetTimes)

        #invert this window to get the fundamental frequency,with the n/n-1
        #correction to account for the difference between the time window and
        #time period of the signal (WE REMOVE THIS FOR NOW)
        f0 = 1.0 / window

        #subtract the mean from the inputData,BEFORE YOU ZERO-PAD IT!
        inputDataCentered = np.concatenate((inputData - np.mean(inputData), np.zeros(len(inputSetTimes))))

        #calculate a variance for the centered data
        varianceInputPoints = np.var(inputDataCentered, ddof=1)

        #define the frequency step as HALF the fundamental frequency in order
        #to zero-pad and get an evenly spaced mesh
        freqStep = 0.5 * f0

        #get the list of frequencies
        freq = np.linspace(0.5 * f0, n * UpperFrequencyFactor * 0.5 * f0, int(n * UpperFrequencyFactor))

        #calculate the inverse autocorrelation
        inverseAuto = 1.0 / (2.0 * varianceInputPoints) * np.array(tuple(map(lambda f: ampSquaredNormed(np.cos, f, inputTimesNormed, inputDataCentered) + ampSquaredNormed(np.sin, f, inputTimesNormed, inputDataCentered), list(freq))))
    
        #return: 1) the list of frequencies, 2) the correspoinding list of inverse autocovariances
        return np.transpose(np.vstack((freq, inverseAuto)))

    inputInverseAuto = InverseAutocovariance(inputTimes[np.isnan(inputData) == False], inputData[np.isnan(inputData) == False], inputSetTimes, UpperFrequencyFactor = UpperFrequencyFactor)

    #create the amplitude spectrum from the input data:
    #add a zero at the first element to make the DFT work, sample only half the
    #points because we have oversampled by 2 in the inverseAutocovariance
    inverseAmplitudes = np.concatenate(([0], inputInverseAuto[:np.int(inputInverseAuto.shape[0] / 2), 1]))

    #do the DCT-III transform:
    autoCorrs = scipy.fftpack.dct(inverseAmplitudes, type=3, norm='ortho')

    #divide everything by a normalization factor so that the autocorrelation at lag 0 = 1
    #make sure we are only returning autocorrelations for the lags we can rely on, i.e.  for up to N/2 time points
    values = autoCorrs[:np.int(np.floor(0.5 * len(autoCorrs)))] / autoCorrs[0]

    return np.vstack((inputSetTimes[:len(values)], values))

def pAutocorrelation(args):

    """Wrapper of Autocorrelation function for use with Multiprocessing.

    Parameters:
        args: tuple
            A tuple of arguments in the form (inputTimes, inputData, inputSetTimes):
                inputTimes: 1d numpy.array
                    Times corresponding to provided data points

                inputData: 1d numpy.array
                    Data points

                inputSetTimes: 1d numpy.array
                    A complete set of all possible N times during which data could have been collected.

    Returns:
        2d numpy.array
            Array of time lags with corresponding autocorrelations

    Usage:
        result = pAutocorrelation((inputTimes, inputData, inputSetTimes))
    """

    inputTimes, inputData, inputSetTimes = args
    
    return autocorrelation(inputTimes, inputData, inputSetTimes)

def getSpikes(inputData, func, cutoffs):

    """Get sorted index of signals with statistically significant spikes,
    i.e. those that pass the provided cutoff.

    Parameters:
        inputData: 2d numpy.array
            Data points where rows are normalized signals

        func: function
            Function np.max or np.min

        cutoffs: dictionary
            A dictionary of cutoff values

    Returns:
        list
            Index of data with statistically significant spikes

    Usage:
        index = getSpikes(inputData, np.max, cutoffs)
    """

    data = inputData.copy()
    counts_non_missing = np.sum(~np.isnan(data), axis=1)
    data[np.isnan(data)] = 0.

    spikesIndex = []

    for i in list(range(data.shape[1]+1)):
        ipos = np.where(counts_non_missing==i)[0]
        if len(data[ipos])>0:
            points = func(data[ipos], axis=1)

            if i in cutoffs.keys():
                spikesIndex.extend(ipos[np.where((points>cutoffs[i][0]) | (points<cutoffs[i][1]))[0]])
            else:
                print('Cuttoff for %s non-missing points is missing'%(i))

    return sorted(spikesIndex)

def LombScargle(inputTimes, inputData, inputSetTimes, FrequenciesOnly=False,NormalizeIntensities=False,OversamplingRate=1,UpperFrequencyFactor=1):

    """Calculate Lomb-Scargle periodogram.

    Parameters:
        inputTimes: 1d numpy.array 
            Times corresponding to provided data points (1D array of floats)

        inputData: 1d numpy.array 
            Data points

        inputSetTimes: 1d numpy.array 
            A complete set of all possible N times during which data could have been collected

        FrequenciesOnly: boolean, Default False
            Return frequencies only

        NormalizeIntensities: boolean, Default False
            Normalize intensities to unity

        OversamplingRate: int, Default 1
            Oversampling rate

        UpperFrequencyFactor: float, Default 1
            Upper frequency factor

    Returns:
        2d numpy.array
            Periodogram with a list of frequencies.

    Usage:
        pgram = LombScargle(inputTimes, inputData, inputSetTimes)
    """

    #adjust inputTimes starting point w.r.t.dataset times
    inputTimesNormed = inputTimes - inputSetTimes[0]

    #calculate the number of timepoints in the overall set
    n = len(inputSetTimes)

    #calculate the time window of observation
    window = np.max(inputSetTimes) - np.min(inputSetTimes)

    #invert this window to get the fundamental frequency, with the n/n-1
    #correction to account for the difference between the time window and time
    #period of the signal (WE ARE,FOR NOW,NOT INCLUDING THIS!)
    f0 = n / ((n - 1) * window)

    #subtract the mean from the inputData
    inputDataCentered = inputData - np.mean(inputData)

    #calculate a variance for the centered data
    varianceInputPoints = np.var(inputDataCentered, ddof=1)

    #define a frequency step
    freqStep = 1 / (OversamplingRate * (np.floor(n / 2) - 1)) * (n / 2 * UpperFrequencyFactor - 1) * f0

    #get the list of frequencies, adjusting both the lower frequency (to equal
    #f0 0-effectively a lowpass filter) and the upper cutoff Nyquist by the upper factor specified
    freq = np.linspace(f0, n / 2 * UpperFrequencyFactor * f0, int(f0 * (n / 2 * UpperFrequencyFactor) / freqStep))

    if FrequenciesOnly:
        return freq

    #get the periodogram
    periodogram = 1.0 / (2.0 * varianceInputPoints) * np.array(tuple(map(lambda f: chop(ampSquaredNormed(np.cos, f, inputTimesNormed, inputDataCentered)) + chop(ampSquaredNormed(np.sin, f, inputTimesNormed, inputDataCentered)), list(freq))))
    
    #the function finally returns:1) the list of frequencies, 2) the
    #corresponding list of Lomb-Scargle spectra
    if NormalizeIntensities:
        periodogram = periodogram / np.sqrt(np.dot(periodogram,periodogram))

    returning = np.vstack((freq, periodogram))

    return returning

def pLombScargle(args):

    """Wrapper of LombScargle function for use with Multiprocessing.

    Parameters:
        args: 
            A tuple of arguments in the form (inputTimes, inputData, inputSetTimes):
                inputTimes: 1d numpy.array
                    Times corresponding to provided data points

                inputData: 1d numpy.array
                    Data points

                inputSetTimes: 1d numpy.array
                    A complete set of all possible N times during which data could have been collected.

    Returns:
        2d numpy.array
            Array of frequencies with corresponding intensities

    Usage:
        result = pLombScargle((inputTimes, inputData, inputSetTimes))
    """

    inputTimes, inputData, inputSetTimes = args
    
    return LombScargle(inputTimes, inputData, inputSetTimes)

def getAutocorrelationsOfData(params):

    """Calculate autocorrelation using Lomb-Scargle Autocorrelation.
    NOTE: there should be already no missing or non-numeric points in the input Series or Dataframe

    Parameters:
        params: tuple
            A tuple of parameters in the form (df_data, setAllInputTimes), where
            df_data is a pandas Series or Dataframe, 
            setAllInputTimes is a complete set of all possible N times during which data could have been collected.

    Returns:
        2d numpy.array
            Array of autocorrelations of data.

    Usage:
        result  = autocorrelation(df_data, setAllInputTimes)
    """

    df, setAllInputTimes = params

    if isinstance(df, pd.Series):

        return autocorrelation(df.index.values, df.values, setAllInputTimes)

    elif isinstance(df, pd.DataFrame):
        listOfAutocorrelations = []

        for timeSeriesIndex in df.index:
            listOfAutocorrelations.append(autocorrelation(df.loc[timeSeriesIndex].index.values, df.loc[timeSeriesIndex].values, setAllInputTimes))

        return np.vstack(listOfAutocorrelations)

    print('Warning: Input data type unrecognized: use <pandas.Series> or <pandas.DataFrame>')

    return None

def metricCommonEuclidean(u,v):

    """Metric to calculate 'euclidean' distance between vectors u and v 
    using only common non-missing points (not NaNs).

    Parameters:
        u: 1d numpy.array
            Numpy 1-D array

        v: 1d numpy.array
            Numpy 1-D array

    Returns:
        float
            Measure of the distance between u and v

    Usage:
        dist = metricCommonEuclidean(u,v)
    """

    where_common = (~np.isnan(u)) * (~np.isnan(v))

    return np.sqrt(((u[where_common] - v[where_common]) ** 2).sum())

def chop(expr, tolerance=1e-10):

    """Equivalent of Mathematica.Chop Function.

    Parameters:
        expr: float, tuple, list or numpy.array
            A number or a pyhton sequence of numbers

        tolerance: 
            Default is the same as in Mathematica

    Returns:
        Chopped data

    Usage
        data = chop(data)
    """
        
    if isinstance(expr, (list, tuple, np.ndarray)):

        expr_copy = np.copy(expr)
        expr_copy[np.abs(expr) < tolerance] = 0

    else:
        expr_copy = 0 if expr < tolerance else expr

    return expr_copy
